from minorminer import busclique
from minorminer.utils import verify_embedding
import unittest, random, itertools, dwave_networkx as dnx, networkx as nx, os

def subgraph_node_yield(g, q):
    """
    returns a subgraph of g produced by removing nodes with probability 1-q
    """
    h = g.copy()
    h.remove_nodes_from([v for v in h if random.random() > q])
    return h

def subgraph_edge_yield(g, p):
    """
    returns a subgraph of g produced by removing edges with probability 1-p.
    """
    h = g.copy()
    h.remove_edges_from([e for e in h.edges() if random.random() > p])
    return h

def subgraph_edge_yield_few_bad(g, p, b):
    """
    returns a subgraph of g produced by removing b internal edges and 
    odd/external edges with probability 1-p.

    g must be a chimera graph or a pegasus graph with coordinate labels
    """
    if g.graph['family'] == 'pegasus' or g.graph['family'] == 'zephyr':
        def icpl(p, q):
            return p[0] != q[0]
    elif g.graph['family'] == 'chimera':
        def icpl(p, q):
            return p[2] != q[2]
    else:
        raise ValueError("g is not a pegasus, zephyr or chimera")

    h = g.copy()
    iedges, oedges = [], []
    for e in h.edges():
        if icpl(*e):
            iedges.append(e)
        elif random.random() > p:
            oedges.append(e)

    if len(iedges) >= b:
        iedges = random.sample(iedges, b)
    else:
        iedges = []

    h.remove_edges_from(itertools.chain(iedges, oedges))
    return h

def relabel(g, node_label, edge_label, *args, **kwargs):
    nodes = node_label(g)
    edges = edge_label(g.edges())

    if g.graph['family'] == 'pegasus':
        f = dnx.pegasus_graph
    elif g.graph['family'] == 'chimera':
        f = dnx.chimera_graph
    elif g.graph['family'] == 'zephyr':
        f = dnx.zephyr_graph
    else:
        raise ValueError("g is not a pegasus, zephyr or chimera")

    return f(*args, edge_list = edges, node_list = nodes, **kwargs)

def max_chainlength(emb):
    if emb:
        return max(len(c) for c in emb.values())

class TestBusclique(unittest.TestCase):
    def __init__(self, *args, **kwargs):
        super(TestBusclique, self).__init__(*args, **kwargs)

        busclique.busgraph_cache.clear_all_caches()

        self.c16 = dnx.chimera_graph(16)
        self.p16 = dnx.pegasus_graph(16)
        self.z6 = dnx.zephyr_graph(6)
        self.c16c = dnx.chimera_graph(16, coordinates=True, data = False)
        self.c428 = dnx.chimera_graph(4, n = 2, t = 8)
        self.c248 = dnx.chimera_graph(2, n = 4, t = 8)
        self.c42A = dnx.chimera_graph(4, n = 2, t = 9)
        c4c_0 = subgraph_node_yield(dnx.chimera_graph(4, coordinates = True), .95)
        p4c_0 = subgraph_node_yield(dnx.pegasus_graph(4, coordinates = True), .95)
        z4c_0 = subgraph_node_yield(dnx.zephyr_graph(4, coordinates = True), .95)
        c4c = [c4c_0,
               subgraph_edge_yield(c4c_0, .95),
               subgraph_edge_yield_few_bad(c4c_0, .95, 6)]
        p4c = [p4c_0,
               subgraph_edge_yield(p4c_0, .95),
               subgraph_edge_yield_few_bad(p4c_0, .95, 6)]
        z4c = [z4c_0,
               subgraph_edge_yield(z4c_0, .95),
               subgraph_edge_yield_few_bad(z4c_0, .95, 6)]

        p4coords = dnx.pegasus_coordinates(4)
        c4coords = dnx.chimera_coordinates(4, 4, 4)
        z4coords = dnx.zephyr_coordinates(4)

        c4 = [relabel(c,
                      c4coords.iter_chimera_to_linear,
                      c4coords.iter_chimera_to_linear_pairs,
                      4) for c in c4c]

        p4 = [relabel(p,
                      p4coords.iter_pegasus_to_linear,
                      p4coords.iter_pegasus_to_linear_pairs,
                      4) for p in p4c]

        p4n = [relabel(p,
                       p4coords.iter_pegasus_to_nice,
                       p4coords.iter_pegasus_to_nice_pairs,
                       4, nice_coordinates = True) for p in p4c]

        z4 = [relabel(z,
                      z4coords.iter_zephyr_to_linear,
                      z4coords.iter_zephyr_to_linear_pairs,
                      4) for z in z4c]


        self.c4, self.c4_nd, self.c4_d = list(zip(c4, c4c))
        self.p4, self.p4_nd, self.p4_d = list(zip(p4, p4c, p4n))
        self.z4, self.z4_nd, self.z4_d = list(zip(z4, z4c))

    def test_p16(self):
        def reconstruct(nodes):
            return dnx.pegasus_graph(16, node_list = nodes)
        self.run_battery('p16', self.p16, reconstruct, 180, 17, 172, 15)

    def test_c16(self):
        def reconstruct(nodes):
            return dnx.chimera_graph(16, node_list = nodes)
        self.run_battery('c16', self.c16, reconstruct, 64, 17, 64, 16)

    def run_battery(self, name, g, reconstruct, cliquesize, cliquelength,
                          bicliquesize, bicliquelength, test_nocache = True):
        labels = g.graph['labels']

        clique_init_ok = False
        test = self.clique_battery(g, reconstruct, test_nocache = test_nocache)
        with self.subTest(msg=f"clique_battery init"):
            emb0, size, cl = next(test)
            s = nx.complete_graph(size)
            verify_embedding(emb0, s, g)
            self.assertEqual(size, cliquesize)
            self.assertEqual(cl, cliquelength)
            clique_init_ok = True
        if clique_init_ok:
            for i, (h, emb_f, kind, check_cl) in enumerate(test):
                with self.subTest(msg=f"clique_battery({name}, {labels}, {kind})"):
                    emb = emb_f()
                    verify_embedding(emb, s, h)
                    if check_cl:
                        self.assertEqual(max_chainlength(emb), cl)

        biclique_init_ok = False
        test = self.biclique_battery(g, reconstruct)
        with self.subTest(msg=f"biclique_battery init"):
            emb0, size, cl = next(test)
            s = nx.complete_bipartite_graph(size, size)
            verify_embedding(emb0, s, g)
            if bicliquelength is not None:
                self.assertEqual(cl, bicliquelength)
            biclique_init_ok = True
        if biclique_init_ok:
            for i, (h, emb_f, kind) in enumerate(test):
                with self.subTest(msg=f"biclique_battery({name}, {labels}, {kind})"):
                    emb = emb_f()
                    verify_embedding(emb, s, h)
                    if bicliquelength is not None:
                        self.assertEqual(max_chainlength(emb), bicliquelength)

    def clique_battery(self, g, reconstruct, test_nocache = True):
        bgcg = busclique.busgraph_cache(g)
        emb0 = bgcg.largest_clique()
        size = len(emb0)
        cl = max_chainlength(emb0)
        N = range(size)
        yield emb0, size, cl
        yield g, lambda:emb0, 'g:bcg.lc', True
        yield g, lambda:bgcg.find_clique_embedding(size), 'g:bcg.fce', True
        yield g, lambda:busclique.find_clique_embedding(size, g), 'g:bc.fce', True
        if test_nocache:
            yield (g,
                   lambda:busclique.find_clique_embedding(size, g, use_cache = False),
                   'g:bc.fce,nc', True)
        yield g, lambda:bgcg.largest_clique_by_chainlength(cl), 'g:bc.lcbc', True
        nodes = set(itertools.chain.from_iterable(emb0.values()))
        h = reconstruct(nodes)
        bgch = busclique.busgraph_cache(h)
        yield h, lambda:busclique.find_clique_embedding(N, h), 'h:bc.fce', True
        if test_nocache:
            yield (h, lambda:busclique.find_clique_embedding(N, h, use_cache = False),
                   'h:bc.fce,nc', True)
        yield h, lambda:bgch.largest_clique(), 'h:bgc.lc', True
        yield h, lambda:bgch.find_clique_embedding(N), 'h:bgc.fce', True
        yield h, lambda:bgch.largest_clique_by_chainlength(cl), 'h:bgc.lcbc', True

    def biclique_battery(self, g, reconstruct):
        bgcg = busclique.busgraph_cache(g)
        emb0 = bgcg.largest_balanced_biclique()
        size = len(emb0)//2
        cl = max_chainlength(emb0)
        N = range(size)
        yield emb0, size, cl
        yield g, lambda:emb0, 'bgc.lbb'
        yield (g, lambda:bgcg.find_biclique_embedding(N, range(size, 2*size)),
                  'bgc.fbe,list')
        yield g, lambda:bgcg.find_biclique_embedding(size, size), 'bgc.fbe,ints'

    @classmethod
    def tearDownClass(cls):
        rootdir = busclique.busgraph_cache.cache_rootdir()
        if not os.path.exists(rootdir):
            raise RuntimeError("cache rootdir not found before cleanup")
        busclique.busgraph_cache.clear_all_caches()
        if os.path.exists(rootdir):
            raise RuntimeError("cache rootdir exists after cleanup")

    def test_chimera_weird_sizes(self):
        self.assertRaises(NotImplementedError,
                          busclique.busgraph_cache,
                          self.c42A)

        self.assertRaises(NotImplementedError,
                          busclique.find_clique_embedding,
                          999, self.c42A)

        self.assertRaises(NotImplementedError,
                          busclique.find_clique_embedding,
                          999, self.c42A, use_cache = False)

        def reconstructor(m, n, t):
            return lambda nodes: dnx.chimera_graph(m, n = n, t = t,
                                                   node_list = nodes)
        for g, params in (self.c428, (4, 2, 8)), (self.c248, (2, 4, 8)):
            reconstruct = reconstructor(*params)
            self.run_battery('c%d%d%d'%params, g, reconstruct, 16, 3, 16, None)

    def test_labelings(self):
        def reconstructor(g):
            return lambda nodes: g.subgraph(nodes).copy()

        names = 'c4_nd', 'c4', 'c4_d', 'p4_nd', 'p4', 'p4_d', 'z4_nd', 'z4', 'z4_d'
        nocache = False, True, True, False, True, True, False, True, True
        topos = self.c4_nd, self.c4, self.c4_d, self.p4_nd, self.p4, self.p4_d, self.z4_nd, self.z4, self.z4_d

        for (name, test_nocache, G) in zip(names, nocache, topos):
            g0 = G[0]
            bgc = busclique.busgraph_cache(g0)
            K = bgc.largest_clique()
            skip_K = True
            with self.subTest(msg = f"test_labelings setup for {name} cliques"):
                verify_embedding(K, nx.complete_graph(len(K)), g0)
                skip_K = False

            B = bgc.largest_balanced_biclique()
            skip_B = True
            with self.subTest(msg = f"test_labelings setup for {name} bicliques"):
                b2 = len(B)//2
                verify_embedding(B, nx.complete_bipartite_graph(b2, b2), g0)
                skip_B = False
            if skip_K or skip_B:
                continue
            for g in G:
                self.run_battery(name, g, reconstructor(g), 
                                 len(K), max_chainlength(K),
                                 len(B)//2, None,
                                 test_nocache = test_nocache)

    def test_k4_bug(self):
        edges = [30, 2940], [30, 2955], [45, 2940], [45, 2955], [2940, 2955]
        p = dnx.pegasus_graph(16, edge_list = edges)
        k4 = busclique.find_clique_embedding(4, p)

        self.assertEqual(k4, {})
        
    def test_seed_determinism(self):
        for g in (self.c4_nd[0], self.z4_nd[0], self.p4_nd[0]):
            seed = random.randint(0, 2**32-1)
            with self.subTest(msg=f"seed determinism test: {g.graph['family']}"):
                bgc = busclique.busgraph_cache(g, seed = seed)
                Klarge = bgc.largest_clique()
                nK = len(Klarge)
                Blarge = bgc.largest_balanced_biclique()
                nB = len(Blarge)//2
                Kfamily = [bgc.find_clique_embedding(i) for i in range(nK+1)]
                Bfamily = [bgc.find_biclique_embedding(i,i) for i in range(nB+1)]
                # now, clear the cache and go again
                busclique.busgraph_cache.clear_all_caches()
                bgc = busclique.busgraph_cache(g, seed = seed)
                self.assertEqual(Klarge, bgc.largest_clique())
                self.assertEqual(Blarge, bgc.largest_balanced_biclique())
                self.assertEqual(Kfamily, [bgc.find_clique_embedding(i) for i in range(nK+1)])
                self.assertEqual(Bfamily, [bgc.find_biclique_embedding(i,i) for i in range(nB+1)])

    def test_perfect_z6_clique(self):
        k88 = nx.complete_graph(88)
        bgc = busclique.busgraph_cache(self.z6)
        e88_cache = bgc.largest_clique()
        verify_embedding(e88_cache, k88, self.z6)

        e88_cache2 = busclique.find_clique_embedding(88, self.z6, use_cache=True)
        verify_embedding(e88_cache2, k88, self.z6)

        e88 = busclique.find_clique_embedding(88, self.z6, use_cache=False)
        verify_embedding(e88, k88, self.z6)
        
    def test_perfect_z6_biclique(self):
        k88_88 = nx.complete_bipartite_graph(88, 88)
        bgc = busclique.busgraph_cache(self.z6)
        e88_88_cache = bgc.largest_balanced_biclique()
        verify_embedding(e88_88_cache, k88_88, self.z6)

    def test_z3_determinism(self):
        #first, check an explicit sample found with find_nondeterminism
        z3 = dnx.zephyr_graph(3)
        z3.remove_nodes_from([11, 43, 61, 77, 102, 107, 143, 145, 162, 174, 178, 185, 240, 333])
        z3.remove_edges_from([
            (28, 29), (28, 264), (37, 40), (55, 259), (73, 76), (96, 99), (97, 98), (98, 302),
            (108, 111), (110, 298), (117, 118), (117, 218), (128, 131), (133, 135), (140, 142),
            (158, 161), (160, 281), (169, 171), (187, 190), (207, 208), (212, 215), (235, 236),
            (236, 239), (242, 245), (265, 266), (316, 317), (319, 321),
        ])
        emb0 = busclique.busgraph_cache(z3).largest_clique()
        emb1 = busclique.find_clique_embedding(len(emb0), z3, use_cache=False)
        self.assertEqual(sorted(emb0.values()), sorted(emb1.values()))

        #then, do a quick scan for another example -- this will present as a flaky test if
        #we merely got lucky hammering down the tet above
        self.assertEqual(None, find_nondeterminism('zephyr', 3, 10))

    def test_p4_determinism(self):
        #first, check an explicit sample found with find_nondeterminism
        p4 = dnx.pegasus_graph(4)
        p4.remove_nodes_from([
            38, 56, 69, 81, 92, 97, 107, 112, 113, 207, 209, 212, 240, 258, 264, 277, 281
        ])
        p4.remove_edges_from([
            (12, 13), (14, 17), (27, 28), (28, 219), (29, 234), (32, 255), (37, 189), (44, 225),
            (52, 53), (54, 55), (61, 62), (66, 67), (74, 77), (85, 86), (114, 115), (123, 176),
            (152, 155), (154, 155), (156, 159), (193, 196), (198, 199), (204, 205), (241, 242),
            (249, 250), (266, 269), (271, 272),
        ])
        emb0 = busclique.busgraph_cache(p4).largest_clique()
        emb1 = busclique.find_clique_embedding(len(emb0), p4, use_cache=False)
        self.assertEqual(sorted(emb0.values()), sorted(emb1.values()))

        #then, do a quick scan for another example -- this will present as a flaky test if
        #we merely got lucky hammering down the tet above
        self.assertEqual(None, find_nondeterminism('pegasus', 4, 10))

    def test_c4_determinism(self):
        #first, check an explicit sample found with find_nondeterminism
        c4 = dnx.chimera_graph(4)
        c4.remove_nodes_from([1, 47, 60, 62, 59, 107])
        c4.remove_edges_from([
            (0, 32), (24, 56), (26, 58), (32, 64), (37, 35), (39, 34), (44, 42), (53, 51),
            (64, 69), (87, 95), (96, 103),
        ])
        emb0 = busclique.busgraph_cache(c4).largest_clique()
        emb1 = busclique.find_clique_embedding(len(emb0), c4, use_cache=False)
        self.assertEqual(sorted(emb0.values()), sorted(emb1.values()))

        #then, do a quick scan for another example -- this will present as a flaky test if
        #we merely got lucky hammering down the tet above
        self.assertEqual(None, find_nondeterminism('chimera', 4, 10))

    def test_cache_export_import(self):
        # make sure we're nondeterministic so accidentally re-computing the 
        # cache will be detected
        p4 = self.p4_nd[0]
        bgc = busclique.busgraph_cache(p4, seed=100)
        cliques = [bgc.find_clique_embedding(n) for n in range(len(bgc.largest_clique()))]
        for sparse in (True, False):
            for compress in (True, False):
                with self.subTest(msg=f"sparse: {sparse}, compress: {compress}"):
                    tmp = busclique.busgraph_cache(p4, seed=101)
                    tmp.update_clique_cache([], write_to_disk=True)
                    self.assertEqual(tmp.largest_clique(), {})

                    cache = bgc.get_clique_cache(sparse=sparse, compress=compress)
                    tmp.update_clique_cache(cache, write_to_disk=False)
                    #test loading from memory
                    for n, c in enumerate(cliques):
                        self.assertEqual(set(c.values()), set(tmp.find_clique_embedding(n).values()))

                    tmp.update_clique_cache(cache, write_to_disk=True)

                    tmp = busclique.busgraph_cache(p4, seed=101)
                    #test loading from disk
                    for n, c in enumerate(cliques):
                        self.assertEqual(set(c.values()), set(tmp.find_clique_embedding(n).values()))        

    def test_cache_merge(self):
        p4 = self.p4_nd[0] 
        bgc0 = busclique.busgraph_cache(p4, seed=200)
        bgc1 = busclique.busgraph_cache(p4, seed=201)
        bgc = busclique.busgraph_cache(p4, seed=202)
        bgc.update_clique_cache(bgc0.get_clique_cache(), write_to_disk=False)
        bgc.merge_clique_cache(bgc1, write_to_disk=False)
        for n in range(1, len(bgc.largest_clique())):
            emb = bgc.find_clique_embedding(n)
            emb0 = bgc0.find_clique_embedding(n)
            emb1 = bgc1.find_clique_embedding(n)
            maxlen0 = max(map(len, emb0.values())) if emb0 else 999
            maxlen1 = max(map(len, emb1.values())) if emb1 else 999
            maxlen = max(map(len, emb.values()))
            self.assertEqual(maxlen, min(maxlen0, maxlen1))

        maxlen = max(map(len, bgc.largest_clique().values()))
        for n in range(maxlen):
            emb = bgc.largest_clique_by_chainlength(n)
            emb0 = bgc0.largest_clique_by_chainlength(n)
            emb1 = bgc1.largest_clique_by_chainlength(n)
            self.assertEqual(len(emb), max(len(emb0), len(emb1)))

    def test_cache_insert(self):
        p4 = self.p4_nd[0]
        bgc = busclique.busgraph_cache(p4, seed=300)
        emb = bgc.largest_clique()
        chain = max(emb.values(), key=len)
        
        bgc.insert_clique_embedding([chain], direct=False) # should not overwrite
        emb0 = bgc.find_clique_embedding(1) #should be chainlength 1
        self.assertEqual(len(emb0[0]), 1)
        
        bgc.insert_clique_embedding([chain], direct=True) # should overwrite
        emb1 = bgc.find_clique_embedding(1) #should be the newly-inserted embedding
        self.assertEqual(len(emb1[0]), len(chain))


    def test_mine_clique_embeddings_smoketest(self):
        busclique.mine_clique_embeddings(
            self.p4_nd[0],
            num_seeds=1,
            heuristic_bound=8,
            heuristic_args=dict(tries=1, max_no_improvement=1, chainlength_patience=0, verbose=2),
            heuristic_runs=1,
        )

    def test_regularize_embedding_bug_215(self):
        emb = {0: [206], 1: [127, 130, 134, 242, 133], 2: [129], 3: [126], 4: [209], 5: [179, 135, 136]}
        missing_nodes = (17, 32, 132, 155, 212, 215, 230, 263, 280)
        missing_edges = [
            (8, 11), (9, 162), (13, 231), (15, 198), (22, 216), (25, 222),
            (26, 264), (27, 195), (31, 222), (31, 225), (34, 219), (34, 198),
            (35, 249), (36, 174), (37, 204), (40, 201), (40, 210), (43, 186),
            (45, 163), (46, 216), (48, 189), (49, 243), (54, 181), (54, 174),
            (56, 256), (57, 177), (58, 232), (59, 253), (59, 262), (61, 223),
            (61, 204), (64, 226), (65, 262), (66, 184), (67, 220), (67, 214),
            (70, 217), (70, 220), (71, 253), (72, 163), (77, 229), (80, 253),
            (81, 181), (82, 83), (82, 214), (84, 175), (86, 250), (88, 220),
            (90, 200), (92, 275), (93, 178), (95, 260), (95, 272), (97, 200),
            (97, 205), (98, 236), (99, 169), (100, 224), (103, 203), (103, 209),
            (105, 185), (105, 197), (108, 182), (109, 218), (112, 200),
            (113, 236), (114, 117), (117, 185), (120, 191), (130, 245),
            (131, 278), (131, 281), (134, 251), (164, 167), (172, 173),
            (180, 183), (187, 188), (194, 197), (198, 199), (216, 217),
            (217, 218), (222, 225), (237, 238), (238, 239), (270, 273)
        ]
        p4 = dnx.pegasus_graph(4)
        p4.remove_nodes_from(missing_nodes)
        p4.remove_edges_from(missing_edges)
        busclique._regularize_embedding(p4, emb)                  

    def test_topology_identifier(self):
        perfect_id = '38cad89632b234831d58675091f1bc581c96de65d4b2a0c06c0d94a7f97e21a7'
        p16 = dnx.pegasus_graph(16, coordinates=True)
        bgc = busclique.busgraph_cache(p16)
        self.assertEqual(
            bgc.topology_identifier(),
            perfect_id,
            f'Topology identifier does not match expectation.  If busclique.__cache_version changed, this test needs to be updated.'
        )

        # see minorminer issue #227 on github -- the busclique algorithm does
        # not depend on odd edges and does not include them in its serialization
        odd_edges = []
        relevant_edges = []
        for p, q in p16.edges:
            if p[0] == q[0] and p[-1] == q[-1]:
                odd_edges.append((p, q))
            else:
                relevant_edges.append((p, q))

        # it's actually possible that we will include those edges in the future
        # though -- there are some optimal clique embeddings which utilize odd
        # edges -- let's put in an explicit test to remind ourselves to update
        # this test under that eventuality
        e = random.choice(odd_edges)
        p16.remove_edge(*e)
        bgc_o = busclique.busgraph_cache(p16)
        self.assertEqual(
            bgc_o.topology_identifier(),
            perfect_id,
            f'topology identifier changed after deleting odd edge {e}'
        )

        e = random.choice(relevant_edges)
        p16.remove_edge(*e)
        bgc_e = busclique.busgraph_cache(p16)
        self.assertNotEqual(
            bgc_e.topology_identifier(),
            perfect_id,
            f'topology identifier did not change after removing non-odd edge {e}'
        )

        p16 = dnx.pegasus_graph(16)
        n = random.choice(list(p16.nodes))
        p16.remove_node(n)
        bgc_n = busclique.busgraph_cache(p16)
        self.assertNotEqual(
            bgc_n.topology_identifier(),
            perfect_id,
            f'topology identifier did not change after removing node {n}'
        )


def find_nondeterminism(family, size=4, tries=10):
    if family == 'pegasus':
        generator = dnx.pegasus_graph
        coordinates = dnx.pegasus_coordinates
    elif family == 'zephyr':
        generator = dnx.zephyr_graph
        coordinates = dnx.zephyr_coordinates
    elif family == 'chimera':
        generator = dnx.chimera_graph
        coordinates = dnx.chimera_coordinates

    for i in range(tries):
        H = generator(size, coordinates=True)
        G = subgraph_edge_yield_few_bad(subgraph_node_yield(H, .95), .95, 6)
        emb0 = busclique.busgraph_cache(G, seed = random.randint(0, 2**32-1)).largest_clique()
        emb1 = busclique.find_clique_embedding(len(emb0), G, use_cache=False, seed=random.randint(0, 2**32-1))
        if sorted(emb0.values()) != sorted(emb1.values()):
            coords = coordinates(size)
            bad_nodes = [v for v in H if v not in G]
            node_f = getattr(coords, f"iter_{family}_to_linear");
            edge_f = getattr(coords, f"iter_{family}_to_linear_pairs");
            return (
                list(node_f(bad_nodes)),
                list(edge_f(e for e in H.edges if not G.has_edge(*e) and all(v in G for v in e)))
            )





"""Pack a set of rectangles into a bounding box with minimum area

===========================
Welcome to rectangle-packer
===========================

|PyPI-Versions| |PyPI-Wheel| |Build-Status| |Read-the-Docs| |GitHub-License|

|PyPI-Downloads|

**Primary use:** Given a set of rectangles with fixed orientations,
find a bounding box of minimum area that contains them all with no
overlap.

This project is inspired by Matt Perdeck's blog post `Fast Optimizing
Rectangle Packing Algorithm for Building CSS Sprites`_.

* The latest documentation is available on `Read the Docs`_.
* The source code is available on `GitHub`_.


Installation
============

Install latest version from `PyPI`_:

.. code:: sh

    $ python3 -m pip install rectangle-packer

Or `clone the repository`_ and install with:

.. code:: sh

    $ python3 setup.py install


Basic usage
===========

.. code:: python

    # Import the module
    >>> import rpack

    # Create a bunch of rectangles (width, height)
    >>> sizes = [(58, 206), (231, 176), (35, 113), (46, 109)]

    # Pack
    >>> positions = rpack.pack(sizes)

    # The result will be a list of (x, y) positions:
    >>> positions
    [(0, 0), (58, 0), (289, 0), (289, 113)]

The output positions are the lower left corner coordinates of each
rectangle in the input.

These positions will yield a packing with no overlaps and enclosing
area as small as possible (best effort).

.. note::

    * You must use **positive integers** as rectangle width and height.

    * The module name is **rpack** which is an abbreviation of the package
      name at PyPI (rectangle-packer).

    * The computational time required by :py:func:`rpack.pack` increases by
      the number *and* size of input rectangles.  If this becomes a problem,
      you might need to implement your own `divide-and-conquer algorithm`_.


Examples
========

**Example A:**

.. only:: latex

  .. image:: https://penlect.com/rpack/2.0.1/img/packing_best_10.pdf
     :alt: pack10
     :align: center

.. only:: html

  .. figure:: https://penlect.com/rpack/2.0.1/img/packing_best_10.svg
     :alt: pack10
     :align: center

**Example B:**

.. only:: latex

  .. image:: https://penlect.com/rpack/2.0.1/img/packing_phi.pdf
     :alt: packphi
     :align: center

.. only:: html

  .. figure:: https://penlect.com/rpack/2.0.1/img/packing_phi.svg
     :alt: packphi
     :align: center


**Example C:** Sometimes the input rectangles simply cannot be packed in
a good way. Here is an example of low packing density:

.. only:: latex

  .. image:: https://penlect.com/rpack/2.0.1/img/packing_worst_10.pdf
     :alt: pack10bad
     :align: center

.. only:: html

  .. figure:: https://penlect.com/rpack/2.0.1/img/packing_worst_10.svg
     :alt: pack10bad
     :align: center


**Example D:** The image below is contributed by Paul Brodersen, and
illustrates a solution to a problem discussed at stackoverflow_.

.. image:: https://i.stack.imgur.com/kLat8.png
    :alt: PaulBrodersenExampleImage
    :align: center
    :width: 450px


.. _Read the Docs: https://rectangle-packer.readthedocs.io/en/latest/
.. _GitHub: https://github.com/Penlect/rectangle-packer
.. _`Fast Optimizing Rectangle Packing Algorithm for Building CSS Sprites`: http://www.codeproject.com/Articles/210979/Fast-optimizing-rectangle-packing-algorithm-for-bu
.. _`clone the repository`: https://github.com/Penlect/rectangle-packer
.. _stackoverflow: https://stackoverflow.com/a/53156709/2912349
.. _PyPI: https://pypi.org/project/rectangle-packer/
..  _`divide-and-conquer algorithm`: https://en.wikipedia.org/wiki/Divide-and-conquer_algorithm

.. |PyPI-Versions| image:: https://img.shields.io/pypi/pyversions/rectangle-packer.svg
   :target: https://pypi.org/project/rectangle-packer

.. |PyPI-Wheel| image:: https://img.shields.io/pypi/wheel/rectangle-packer.svg
   :target: https://pypi.org/project/rectangle-packer

.. |Build-Status| image:: https://travis-ci.com/Penlect/rectangle-packer.svg?branch=master
   :target: https://travis-ci.com/Penlect/rectangle-packer

.. |Read-the-Docs| image:: https://img.shields.io/readthedocs/rectangle-packer.svg
   :target: https://rectangle-packer.readthedocs.io/en/latest

.. |GitHub-License| image:: https://img.shields.io/github/license/Penlect/rectangle-packer.svg
   :target: https://raw.githubusercontent.com/Penlect/rectangle-packer/travis/LICENSE.md

.. |PyPI-Downloads| image:: https://img.shields.io/pypi/dm/rectangle-packer.svg
   :target: https://pypi.org/project/rectangle-packer
"""

# Module metadata
__author__ = 'Daniel Andersson'
__maintainer__ = __author__
__email__ = 'daniel.4ndersson@gmail.com'
__contact__ = __email__
__copyright__ = 'Copyright (c) 2017 Daniel Andersson'
__license__ = 'MIT'
__url__ = 'https://github.com/Penlect/rectangle-packer'
__version__ = '2.0.2'

# Built-in
from typing import Iterable, Tuple, List

# Extension modules
from minorminer._extern.rpack._core import (
    pack as _pack,
    PackingImpossibleError,
    bbox_size,
    packing_density,
    overlapping
)

enclosing_size = bbox_size


def pack(sizes: Iterable[Tuple[int, int]],
         max_width=None, max_height=None) -> List[Tuple[int, int]]:
    """Pack rectangles into a bounding box with minimal area.

    The result is returned as a list of coordinates "(x, y)", which
    specifices the location of each corresponding input rectangle's
    lower left corner.

    The helper function :py:func:`bbox_size` can be used to compute
    the width and height of the resulting bounding box.  And
    :py:func:`packing_density` can be used to evaluate the packing
    quality.

    The algorithm will sort the input in different ways internally so
    there is no need to sort ``sizes`` in advance.

    The GIL is released when C-intensive code is running.  Execution
    time increases by the number *and* size of input rectangles.  If
    this becomes a problem, you might need to implement your own
    `divide-and-conquer algorithm`_.

    **Example**::

        # Import the module
        >>> import rpack

        # Create a bunch of rectangles (width, height)
        >>> sizes = [(58, 206), (231, 176), (35, 113), (46, 109)]

        # Pack
        >>> positions = rpack.pack(sizes)

        # The result will be a list of (x, y) positions:
        >>> positions
        [(0, 0), (58, 0), (289, 0), (289, 113)]

    ..  _`divide-and-conquer algorithm`: https://en.wikipedia.org/wiki/Divide-and-conquer_algorithm

    :param sizes: "(width, height)" of the rectangles to pack.  *Note:
        integer values only!*
    :type sizes: Iterable[Tuple[int, int]]

    :param max_width: Force the enclosing rectangle to not exceed a
        maximum width.  If not possible,
        :py:exc:`rpack.PackingImpossibleError` will be raised.
    :type max_width: Union[None, int]

    :param max_height: Force the enclosing rectangle to not exceed a
        maximum height.  If not possible,
        :py:exc:`rpack.PackingImpossibleError` will be raised.
    :type max_height: Union[None, int]

    :return: List of positions (x, y) of the input rectangles.
    :rtype: List[Tuple[int, int]]
    """
    if max_width is not None and not isinstance(max_width, int):
        raise TypeError("max_width must be an integer")
    if max_height is not None and not isinstance(max_height, int):
        raise TypeError("max_height must be an integer")
    if not isinstance(sizes, list):
        sizes = list(sizes)
    return _pack(sizes, max_width or -1, max_height or -1)



# Copyright 2024 D-Wave
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
r"""Determining embedding feasibility is NP-complete.
Efficient methods and heuristics are collected together in this module,
that provide necessary conditions and bounds on feasibility.
This includes determination of a lower bound on the (square)
lattice size required for chimera, pegasus and zephyr dwave_networkx graph
types, and a feasibility filter for general graphs based on the degree
distributions of target and source graphs.
"""
import dwave_networkx as dnx
import networkx as nx
import numpy as np
from collections import Counter
from typing import Optional


def embedding_feasibility_filter(
    S: nx.Graph, T: nx.Graph, one_to_one: bool = False
) -> bool:
    """Feasibility filter for embedding.

    Determines if the source graph ``S`` can be subgraph-embedded onto the target
    graph ``T`` based on their degree distributions. Returns ``False`` if embedding
    is definitely infeasible; otherwise, returns ``True``. False positives are
    permissible because deciding the graph isomorphism problem is NP-complete,
    and this filter is designed to be efficient.

    The degree distribution test is a heuristic; stronger filters are possible
    by exploiting additional graph structure.

    Args:
        S: The source graph to embed.
        T: The target graph in which to embed.
        one_to_one: If True, only 1-to-1 (subgraph) embeddings are allowed.
            Defaults to False, permitting minor embeddings.

    Returns:
        bool: ``False`` if subgraph embedding is definitely infeasible, ``True``
        otherwise.
    """
    if T == S:
        return True
    if (
        T.number_of_nodes() < S.number_of_nodes()
        or T.number_of_edges() < S.number_of_edges()
    ):
        return False
    if S.number_of_nodes() == 0:
        return True
    elif T.number_of_nodes() == 0:
        return False

    S_degree = np.sort([S.degree[n] for n in S.nodes()])
    T_degree = np.sort([T.degree[n] for n in T.nodes()])

    if np.any(T_degree[-len(S_degree) :] < S_degree):
        if one_to_one or T_degree[-1] <= 2:
            # Too many high degree nodes in S
            return False
        # Attempt minor embed (enhance T degrees)
        # Minor embedding feasibility reduces to bin packing when
        # considering a best target case knowing only the degree
        # distribution. In general feasibility is NP-complete, a cheap
        # marginal degree distribution filter is used.

        # We can eliminate nodes of equal degree assuming best case:
        ResidualCounts = Counter(T_degree)
        ResidualCounts.subtract(Counter(S_degree))

        # Target nodes of degree x <=2 are only of use for minor
        # embedding source nodes of degree <=x:
        for kS in range(3):
            if ResidualCounts[kS] < 0:
                ResidualCounts[kS + 1] += ResidualCounts[kS]
            ResidualCounts[kS] = 0

        if all(v > 0 for v in ResidualCounts.values()):
            return True

        nT_auxiliary = sum(ResidualCounts.values())
        if nT_auxiliary < 0:  # extra available to form chains
            return False

        # In best case all target nodes have degree kTmax, and chains
        # are trees. To cover degree k in S requires n auxiliary target
        # nodes such that kTmax + n(kTmax-2) >= k
        kTmax = np.max([k for k, v in ResidualCounts.items() if v > 0])
        min_auxiliary_necessary = sum(
            [
                -v * np.ceil((k - kTmax) / (kTmax - 2))
                for k, v in ResidualCounts.items()
                if v < 0
            ]
        )
        return min_auxiliary_necessary <= nT_auxiliary

    else:
        return True


def lattice_size_lower_bound(
    S: nx.Graph,
    T: nx.Graph = None,
    topology: str = None,
    t: int = None,
    one_to_one: bool = False,
) -> Optional[int]:
    """Returns a lower bound on the size necessary for embedding.

    The lattice size is the parameter ``m`` of a dwave_networkx graph, also
    called number of rows. The function returns a lower bound (necessary but
    not sufficient for embedding) using efficiently established graph
    properties such as the number of nodes, number of edges, node-degree
    distribution, and two-colorability

    Args:
        S: The source graph to embed.
        T: The target graph in which to embed. The graph must be of type
            'zephyr', 'pegasus', or 'chimera' and constructed by dwave_networkx.
        topology: The topology ('chimera', 'pegasus', or 'zephyr'). This is
            inferred from `T` by default. Any set value must be consistent with
            ``T`` (if ``T`` is not None).
        t: The tile parameter, relevant for 'zephyr' and 'chimera' topologies.
            Inferred from ``T`` by default. Any set value must be consistent with
            ``T`` (if ``T`` is not None).
        one_to_one: True if a subgraph embedding is assumed, False for general
            embeddings.

    Raises:
        ValueError: If ``T`` and ``topology`` are inconsistent, or if ``T`` and ``t`` are inconsistent.
        ValueError: If neither ``T`` nor ``topology`` is specified.
        ValueError: If ``S`` cannot be embedded in ``T`` or the specified topology.

    Returns:
        float: Minimum ``lattice_size`` for embedding to be feasible. Returns
        None if embedding for any number of graph rows is infeasible.
    """
    # Comment, similar bounds are possible allowing for minor embeddings,
    # the degree distribution is bounded as chain length increases, whilst
    # the number of nodes is depleted.
    # This could be a possible feature expansion.
    if T is not None:
        if embedding_feasibility_filter(S, T, one_to_one) is False:
            return None

        if topology is None:
            topology = T.graph.get("family")
        elif topology != T.graph.get("family"):
            raise ValueError("Arguments T and topology are inconsistent")

        if t is None:
            t = T.graph["tile"]
    else:
        if topology is None:
            raise ValueError("T or topology must be specified")

        if t is None:
            t = 4
        max_degrees = {"chimera": 2 + 2 * t, "pegasus": 15, "zephyr": 4 + 4 * t}
        max_source_degree = max(d for _, d in S.degree)

        if max_source_degree > max_degrees[topology]:
            return None

    N = S.number_of_nodes()
    if topology == "chimera":
        # Two colorability is necessary, and cheap to check
        if any(
            c > 2
            for c in nx.algorithms.coloring.greedy_coloring.greedy_color(
                S, strategy="connected_sequential_bfs"
            ).values()
        ):
            return None

        def generator(lattice_size):
            return dnx.chimera_graph(m=lattice_size, n=lattice_size, t=t)

        # A lower bound based on number of variables N = m*n*2*t
        lattice_size = np.ceil(np.sqrt(N / 4 / t))

    elif topology == "pegasus":

        def generator(lattice_size):
            return dnx.pegasus_graph(m=lattice_size)

        # A lower bound based on number of variables N = (m*24-8)*(m-1)
        lattice_size = np.ceil(1 / 12 * (8 + np.sqrt(6 * N + 16)))

    elif topology == "zephyr":

        def generator(lattice_size):
            return dnx.zephyr_graph(m=lattice_size, t=t)

        # A lower bound based on number of variables N = (2m+1)*m*4*t
        lattice_size = np.ceil((np.sqrt(2 * N / t + 1) - 1) / 4)

    else:
        raise ValueError(
            "source graphs must be a graph constructed by "
            "dwave_networkx as chimera, pegasus or zephyr type"
        )

    # Evaluate tile feasibility (defect free subgraphs)
    lattice_size = round(lattice_size)
    tile = generator(lattice_size=lattice_size)
    while embedding_feasibility_filter(S, tile, one_to_one) is False:
        lattice_size += 1
        tile = generator(lattice_size=lattice_size)
    return lattice_size



from minorminer import subgraph
from minorminer.utils import verify_embedding
import unittest, random, itertools, dwave_networkx as dnx, networkx as nx, os

def verify_homomorphism(emb, source, target, locallyinjective = False):
    for s in source:
        if s not in emb:
            raise RuntimeError("homomorphism node fail")

        nbrs = {emb[v] for v in source[s]}
        if not nbrs.issubset(target[emb[s]]):
            raise RuntimeError("homomorphism edge fail")

        if locallyinjective and len(source[s]) + 1 > len(nbrs | {emb[s]}):
            raise RuntimeError("homomorphism locality fail")

def as_embedding(emb):
    return {k: (v,) for k, v in emb.items()}

class TestSubgraph(unittest.TestCase):
    def verify_node_coloring(self, emb, source, node_labels):
        for v in source:
           self.assertEqual(node_labels[0].get(v), node_labels[1].get(emb[v]))

    def verify_edge_coloring(self, emb, source, edge_labels):
        for uv in source.edges:
            u, v = uv
            pq = emb[u], emb[v]
            self.assertEqual(edge_labels[0].get(uv), edge_labels[1].get(pq))
            self.assertEqual(edge_labels[0].get(uv[::-1]), edge_labels[1].get(pq[::-1]))

    def test_smoketest(self):
        b = nx.complete_bipartite_graph(4, 4)
        c = nx.cubical_graph()
        emb = subgraph.find_subgraph(c, b)
        verify_embedding(as_embedding(emb), c, b)

    def test_as_embedding(self):
        b = nx.complete_bipartite_graph(4, 4)
        c = nx.cubical_graph()
        emb = subgraph.find_subgraph(c, b, as_embedding=True)
        verify_embedding(emb, c, b)

    def test_node_colors(self):
        path_5 = nx.path_graph(5)
        path_10 = nx.path_graph('qrstuvwxyz')
        node_labels = {0: "start", 4:"end"}, {'v': "start", 'r': "end"}
        emb = subgraph.find_subgraph(path_5, path_10, node_labels=node_labels)
        self.verify_node_coloring(emb, path_5, node_labels)
        verify_embedding(as_embedding(emb), path_5, path_10)

        #do it again with generators instead of graphs
        emb = subgraph.find_subgraph(iter(path_5.edges), iter(path_10.edges), node_labels=node_labels)
        self.verify_node_coloring(emb, path_5, node_labels)
        verify_embedding(as_embedding(emb), path_5, path_10)

        #and again with edge lists
        emb = subgraph.find_subgraph(list(path_5.edges), list(path_10.edges), node_labels=node_labels)
        self.verify_node_coloring(emb, path_5, node_labels)
        verify_embedding(as_embedding(emb), path_5, path_10)

        node_labels = {0: "start", 4:"end"}, {'v': "start", 'q': "end"} #impossible
        emb = subgraph.find_subgraph(path_5, path_10, node_labels=node_labels)
        self.assertEqual(emb, {})

    def test_edge_colors(self):
        path_5 = nx.path_graph('vwxyz')
        path_10 = nx.path_graph(10)

        #this is a directed edge thing
        edge_labels = {('v','w'): "start"}, {(1, 0): "start"} #impossible
        emb = subgraph.find_subgraph(path_5, path_10, edge_labels=edge_labels)
        self.assertEqual(emb, {})

        #try that again but undirected
        edge_labels = {('v','w'): "start", ('w', 'v'): "start"}, {(0,1): "start", (1, 0): "start"}
        emb = subgraph.find_subgraph(path_5, path_10, edge_labels=edge_labels)
        self.verify_edge_coloring(emb, path_5, edge_labels)
        verify_embedding(as_embedding(emb), path_5, path_10)

        #now with a solvable directed problem
        edge_labels = {('v','w'): "start"}, {(7, 6): "start"}
        emb = subgraph.find_subgraph(path_5, path_10, edge_labels=edge_labels)
        self.verify_edge_coloring(emb, path_5, edge_labels)
        verify_embedding(as_embedding(emb), path_5, path_10)
            
    def test_timeout(self):
        source = dnx.chimera_graph(8)
        target = dnx.chimera_graph(15, coordinates=True)
        #pop out a vertex from the central tile to make a minimally-impossible
        #problem (no fully-yielded 8x8s) that GSS know how to reason about
        target.remove_node((7,7,0,0)) 
        emb = subgraph.find_subgraph(source, target, timeout=2)
        self.assertEqual(emb, {})

    def test_noninjective(self):
        # find a 2-coloring of Chimera!
        source = dnx.chimera_graph(4)
        target = nx.path_graph(2)
        emb = subgraph.find_subgraph(source, target, injectivity='noninjective')
        verify_homomorphism(emb, source, target)

        # find a 4-coloring of zephyr!
        source = dnx.zephyr_graph(4, t=1)
        target = nx.complete_graph(4)
        emb = subgraph.find_subgraph(source, target, injectivity='noninjective')
        verify_homomorphism(emb, source, target)

        # but not a 3-coloring!
        target = nx.complete_graph(3)
        emb = subgraph.find_subgraph(source, target, injectivity='noninjective')
        self.assertEqual(emb, {})

    def test_locally_injective(self):
        # find a triple-cover of a 3-cycle by a 9-cycle
        source = nx.cycle_graph('rstuvwxyz')
        target = nx.cycle_graph(3)
        emb = subgraph.find_subgraph(source, target, injectivity='locally injective')
        verify_homomorphism(emb, source, target, locallyinjective=True)

        # can't find a locally injective homomorphism for a 10-cycle into a 3-cycle
        source = nx.cycle_graph(10)
        emb = subgraph.find_subgraph(source, target, injectivity='locally injective')
        self.assertEqual(emb, {})

    def test_isolated_nodes(self):
        source = nx.cycle_graph('rstuvwxyz')
        target = nx.cycle_graph(9)
        source.add_node('a')

        emb = subgraph.find_subgraph(source, target)
        self.assertEqual(emb, {})

        emb = subgraph.find_subgraph(source, target, injectivity='locally injective')
        verify_homomorphism(emb, source, target, locallyinjective=True)

        target.add_node('a')
        emb = subgraph.find_subgraph(source, target, as_embedding=True)
        verify_embedding(emb, source, target)

        target.add_edge('a', 'x')
        emb = subgraph.find_subgraph(source, target, as_embedding=True)
        verify_embedding(emb, source, target)

        #annoying edge case: the source graph is edgeless :eyeroll:
        source = nx.empty_graph(5)
        emb = subgraph.find_subgraph(source, target, as_embedding=True)
        verify_embedding(emb, source, target)

        target = nx.path_graph('ab')
        target.add_node(0)
        emb = subgraph.find_subgraph(source, target, injectivity='locally injective')
        verify_homomorphism(emb, source, target, locallyinjective=True)

    def test_seed(self):
        g = dnx.chimera_graph(3, 3, 1, coordinates=True)
        emb = subgraph.find_subgraph(g, g, seed=54321, as_embedding=True)
        verify_embedding(emb, g, g)

        emb = subgraph.find_subgraph(g, g, seed=random, as_embedding=True)
        verify_embedding(emb, g, g)

        emb = subgraph.find_subgraph(g, g, seed=random.Random(12345), as_embedding=True)
        verify_embedding(emb, g, g)

        #again, with iterable of edges
        emb = subgraph.find_subgraph(iter(g.edges), g, seed=random.Random(12345), as_embedding=True)
        verify_embedding(emb, g, g)

        #again, with list of edges
        emb = subgraph.find_subgraph(list(g.edges), g, seed=random.Random(12345), as_embedding=True)
        verify_embedding(emb, g, g)

    def test_seed_and_colors(self):
        g = nx.petersen_graph()
        label_nodes = random.sample(list(g), 2)
        e0, e1 = random.sample(list(g.edges), 2)
        node_labels = {x: "hello" for x in label_nodes}
        edge_labels = {x: "hello" for x in (e0, e1, e1[::-1])}
        node_labels = node_labels, node_labels
        edge_labels = edge_labels, edge_labels

        emb = subgraph.find_subgraph(g, g, seed=random, node_labels=node_labels)
        verify_embedding(as_embedding(emb), g, g)
        self.verify_node_coloring(emb, g, node_labels)
        
        emb = subgraph.find_subgraph(g, g, seed=random, edge_labels=edge_labels)
        verify_embedding(as_embedding(emb), g, g)
        self.verify_edge_coloring(emb, g, node_labels)

        emb = subgraph.find_subgraph(g, g, seed=random, node_labels=node_labels, edge_labels=edge_labels)
        verify_embedding(as_embedding(emb), g, g)
        self.verify_node_coloring(emb, g, node_labels)
        self.verify_edge_coloring(emb, g, node_labels)

        #again with list / iterable of edges
        emb = subgraph.find_subgraph(list(g.edges), iter(g.edges), seed=random, node_labels=node_labels)
        verify_embedding(as_embedding(emb), g, g)
        self.verify_node_coloring(emb, g, node_labels)
        
        emb = subgraph.find_subgraph(list(g.edges), iter(g.edges), seed=random, edge_labels=edge_labels)
        verify_embedding(as_embedding(emb), g, g)
        self.verify_edge_coloring(emb, g, node_labels)

        emb = subgraph.find_subgraph(list(g.edges), iter(g.edges), seed=random, node_labels=node_labels, edge_labels=edge_labels)
        verify_embedding(as_embedding(emb), g, g)
        self.verify_node_coloring(emb, g, node_labels)
        self.verify_edge_coloring(emb, g, node_labels)




"""Core functions"""

# Built-in
import collections
import math
from typing import Tuple

# Cython
import cython
from libc.stdlib cimport malloc, free, qsort
from libc.limits cimport LONG_MAX
from cpython.mem cimport PyMem_Malloc, PyMem_Free


DEF NO_POSITION = -1
DEF CASE_0 = 0
DEF CASE_1 = 1
DEF CASE_2 = 2
DEF CASE_3 = 3
DEF CASE_4 = 4


class PackingImpossibleError(Exception):
    """Packing rectangles is impossible with imposed restrictions.

    This can happen, for example, if `max_width` is strictly less than
    the widest rectangle given to :py:func:`rpack.pack`.

    If possible, a partial result will be given in the second
    argument, with the positions of packed rectangles up till the
    point of failure.
    """

# The following four functions are used as compare-functions for
# sorting `Rectangle` structs by index, width, height and area.

cdef int rectangle_index_cmp(const void *a, const void *b) noexcept nogil:
    cdef int index_a, index_b
    index_a = (<Rectangle*>a)[0].index
    index_b = (<Rectangle*>b)[0].index
    if index_a < index_b:
        return -1
    elif index_a == index_b:
        return 0
    else:
        return 1


cdef int rectangle_width_cmp(const void *a, const void *b) noexcept nogil:
    cdef int width_a, width_b
    width_a = (<Rectangle*>a)[0].width
    width_b = (<Rectangle*>b)[0].width
    if width_a < width_b:
        return 1
    elif width_a == width_b:
        return rectangle_area_cmp(a, b)
    else:
        return -1


cdef int rectangle_height_cmp(const void *a, const void *b) noexcept nogil:
    cdef int height_a, height_b
    height_a = (<Rectangle*>a)[0].height
    height_b = (<Rectangle*>b)[0].height
    if height_a < height_b:
        return 1
    elif height_a == height_b:
        return rectangle_area_cmp(a, b)
    else:
        return -1


cdef int rectangle_area_cmp(const void *a, const void *b) noexcept nogil:
    cdef int area_a, area_b
    area_a = (<Rectangle*>a)[0].area
    area_b = (<Rectangle*>b)[0].area
    if area_a < area_b:
        return 1
    elif area_a == area_b:
        return 0
    else:
        return -1


cdef class RectangleSet:
    """Container for a set of rectangles to be packed."""

    cdef:
        Rectangle *rectangles
        Py_ssize_t length
        long sum_width
        long sum_height

        long min_width
        long min_height

        long max_width
        long max_height

        long area


    def __cinit__(self, sizes):
        cdef:
            size_t i
            long w, h

        if len(sizes) == 0:
            raise ValueError('Empty sizes')
        self.length = len(sizes)
        self.min_height = LONG_MAX
        self.min_width = LONG_MAX

        # Prepare input
        self.rectangles = NULL
        self.rectangles = <Rectangle *> PyMem_Malloc(<size_t>self.length * sizeof(Rectangle))
        if not self.rectangles:
            raise MemoryError('Rectangles')
        i = 0
        for width, height in sizes:
            if not isinstance(width, int):
                raise TypeError("Rectangle width must be an integer")
            if not isinstance(height, int):
                raise TypeError("Rectangle height must be an integer")
            w = <long>width
            h = <long>height
            if w <= 0:
                raise ValueError("Rectangle width must be positive integer")
            elif h <= 0:
                raise ValueError("Rectangle height must be positive integer")

            self.sum_width += w
            self.sum_height += h

            if self.max_width < w:
                self.max_width = w
            if self.max_height < h:
                self.max_height = h

            if w < self.min_width:
                self.min_width = w
            if h < self.min_height:
                self.min_height = h

            self.area += w*h

            self.rectangles[i].width = w
            self.rectangles[i].height = h
            self.rectangles[i].x = NO_POSITION
            self.rectangles[i].y = NO_POSITION
            self.rectangles[i].index = i
            self.rectangles[i].area = w*h
            self.rectangles[i].wide = (w >= h)
            self.rectangles[i].rotated = False
            i += 1

    def __dealloc__(self):
        # Note: this is called if Exceptions are raised in __cinit__
        if self.rectangles != NULL:
            PyMem_Free(self.rectangles)

    def __iter__(self):
        cdef size_t i
        for i in range(self.length):
            yield self.rectangles[i]

    cdef list positions(self):
        cdef size_t i = 0, end_i = 0
        for i in range(self.length):
            if self.rectangles[i].x == NO_POSITION or \
                   self.rectangles[i].y == NO_POSITION:
                break
            end_i = i + 1
        self.sort_by_index(end_i)
        output = [(self.rectangles[i].x, self.rectangles[i].y)
                  for i in range(end_i)]
        if end_i != self.length:
            raise PackingImpossibleError(f"Partial result", output)
        return output

    cdef bbox_size(self):
        cdef:
            long max_width = 0, max_height = 0
            Rectangle *r
            size_t i
        for i in range(self.length):
            r = &self.rectangles[i]
            if r.x == NO_POSITION or r.y == NO_POSITION:
                break
            if r.width + r.x > max_width:
                max_width = r.width + r.x
            if r.height + r.y > max_height:
                max_height = r.height + r.y
        return max_width, max_height

    cdef void rotate_all(self) nogil:
        cdef size_t i
        cdef Rectangle *r
        for i in range(self.length):
            r = &self.rectangles[i]
            r.width, r.height = r.height, r.width
            r.rotated = not r.rotated
            r.wide = r.width >= r.height
        self.sum_width, self.sum_height = self.sum_height, self.sum_width
        self.min_width, self.min_height = self.min_height, self.min_width
        self.max_width, self.max_height = self.max_height, self.max_width

    cdef void sort_by_index(self, size_t length) nogil:
        qsort(<void*>(self.rectangles), length, sizeof(Rectangle), rectangle_index_cmp)

    cdef void sort_by_width(self) nogil:
        qsort(<void*>(self.rectangles), self.length, sizeof(Rectangle), rectangle_width_cmp)

    cdef void sort_by_height(self) nogil:
        qsort(<void*>(self.rectangles), self.length, sizeof(Rectangle), rectangle_height_cmp)

    cdef void sort_by_area(self) nogil:
        qsort(<void*>(self.rectangles), self.length, sizeof(Rectangle), rectangle_area_cmp)

    cdef void translate(self, long x, long y) nogil:
        cdef size_t i
        cdef Rectangle *r
        for i in range(self.length):
            r = &self.rectangles[i]
            r.x += x
            r.y += y

    cdef void transpose(self) nogil:
        cdef size_t i
        cdef Rectangle *r
        for i in range(self.length):
            r = &self.rectangles[i]
            r.x, r.y = r.y, r.x


cdef class Grid:

    cdef:
        Py_ssize_t length
        CGrid *cgrid

    def __cinit__(self, size_t size, long width=0, long height=0):
        self.cgrid = grid_alloc(size, width, height)
        if not self.cgrid:
            raise MemoryError('Grid')

    def __dealloc__(self):
        if self.cgrid != NULL:
            grid_free(self.cgrid)

    cdef (long, long) search_bbox(
            self, RectangleSet rset, BBoxRestrictions *bbr):
        cdef long status
        if self.cgrid.size + 1 < rset.length:
            raise PackingImpossibleError(
                "Too many rectangles for allocated grid size.", [])
        assert bbr.min_width == rset.max_width
        assert bbr.min_height == rset.max_height
        with nogil:
            status = grid_search_bbox(self.cgrid, rset.rectangles, bbr)
        if status >= 0:
            return self.cgrid.width, self.cgrid.height
        return self.cgrid.width, -self.cgrid.height

    cdef int pack(self, RectangleSet rset, long width, long height) except -1:
        cdef size_t i = 0
        cdef Region reg
        if self.cgrid.size + 1 < rset.length:
            raise PackingImpossibleError(
                "Too many rectangles for allocated grid size.", [])
        with nogil:
            self.cgrid.width = width
            self.cgrid.height = height
            grid_clear(self.cgrid)
            for i in range(rset.length):
                r = &rset.rectangles[i]
                grid_find_region(self.cgrid, r, &reg)
                if reg.col_cell == NULL:
                    r.x = NO_POSITION
                    r.y = NO_POSITION
                    return 1
                r.x = start_pos(reg.col_cell_start)
                r.y = start_pos(reg.row_cell_start)
                grid_split(self.cgrid, &reg)
        return 0


def pack(sizes, long max_width, long max_height):
    """Pack rectangles by testing four different strategies.

    Strategies:

        - Sort by height.

        - Sort by width.

        - Rotate and sort by height.

        - Rotate and sort by width.

    The result of the best one will be returned.

    If ``max_width`` is negative, it will be ignored.  Same for
    ``max_height``.

    If ``max_width`` or ``max_height`` is too restrictive, a
    ``PackingImpossibleError`` exception might be issued.
    """
    cdef:
        long area = LONG_MAX
        long w = 0, h = 0, best_w = 0, best_h = 0
        int case = CASE_0
        BBoxRestrictions bbr

    # Abort early
    n = len(sizes)
    if n == 0:
        return list()

    cdef RectangleSet rset = RectangleSet(sizes)

    if max_width < 0:
        max_width = rset.sum_width
    elif max_width == 0:
        raise PackingImpossibleError("max_width zero", list())
    elif max_width < rset.max_width:
        raise PackingImpossibleError("max_width less than widest rectangle", list())

    if max_height < 0:
        max_height = rset.sum_height
    elif max_height == 0:
        raise PackingImpossibleError("max_height zero", list())
    elif max_height < rset.max_height:
        raise PackingImpossibleError("max_height less than highest rectangle", list())

    grid = Grid(rset.length + 1, 0, 0)
    bbr = BBoxRestrictions(
        min_width=rset.max_width,
        max_width=max_width,
        min_height=rset.max_height,
        max_height=max_height,
        max_area=area
    )
    best_w = max_width
    best_h = max_height

    rset.sort_by_height()
    w, h = grid.search_bbox(rset, &bbr)
    area = w*h
    if 0 < area < bbr.max_area:
        bbr.max_area = area
        best_w = w
        best_h = h
        case = CASE_1

    rset.sort_by_width()
    w, h = grid.search_bbox(rset, &bbr)
    area = w*h
    if 0 < area < bbr.max_area:
        bbr.max_area = area
        best_w = w
        best_h = h
        case = CASE_2

    # Rotated

    rset.rotate_all()
    bbr.min_width = rset.max_width
    bbr.max_width = max_height
    bbr.min_height = rset.max_height
    bbr.max_height = max_width

    w, h = grid.search_bbox(rset, &bbr)
    area = w*h
    if 0 < area < bbr.max_area:
        bbr.max_area = area
        best_w = w
        best_h = h
        case = CASE_3

    rset.sort_by_width()
    w, h = grid.search_bbox(rset, &bbr)
    area = w*h
    if 0 < area < bbr.max_area:
        bbr.max_area = area
        best_w = w
        best_h = h
        case = CASE_4

    # Restore rset to best case
    if case == CASE_0:
        best_w = bbr.max_width
        best_h = bbr.max_height
    elif case == CASE_1:
        rset.rotate_all()
    elif case == CASE_2:
        rset.rotate_all()
        rset.sort_by_width()
    elif case == CASE_3:
        rset.sort_by_height()
    elif case == CASE_4:
        pass

    grid.pack(rset, best_w, best_h)

    # Restore rectangles if rotated
    if case == CASE_0 or case == CASE_3 or case == CASE_4:
        rset.transpose()

    return rset.positions()


def bbox_size(sizes, positions) -> Tuple[int, int]:
    """Return bounding box size (width, height) of packed rectangles.

    Useful for evaluating the result of :py:func:`rpack.pack`.

    Example::

        >>> import rpack

        >>> sizes = [(58, 206), (231, 176), (35, 113), (46, 109)]
        >>> positions = rpack.pack(sizes)

        >>> bbox_size(sizes, positions)
        (335, 222)

    :param sizes: List of rectangle sizes (width, height).
    :type sizes: List[Tuple[int, int]]

    :param positions: List of rectangle positions (x, y).
    :type positions: List[Tuple[int, int]]

    :return: Size (width, height) of bounding box covering rectangles
             having `sizes` and `positions`.
    :rtype: Tuple[int, int]
    """
    cdef long max_width = 0, max_height = 0, w, h, x, y
    for i in range(len(sizes)):
        w, h = sizes[i]
        x, y = positions[i]
        if w + x > max_width:
            max_width = w + x
        if h + y > max_height:
            max_height = h + y
    return max_width, max_height


def packing_density(sizes, positions) -> float:
    """Return packing density of packed rectangles.

    Useful for evaluating the result of :py:func:`rpack.pack`.

    Example::

        >>> import rpack

        >>> sizes = [(58, 206), (231, 176), (35, 113), (46, 109)]
        >>> positions = rpack.pack(sizes)

        >>> packing_density(sizes, positions)
        0.8279279279279279

    :param sizes: List of rectangle sizes (width, height).
    :type sizes: List[Tuple[int, int]]

    :param positions: List of rectangle positions (x, y).
    :type positions: List[Tuple[int, int]]

    :return: Packing density as a fraction in the interval [0, 1],
             where 1 means that the bounding box area equals the sum
             of the areas of the rectangles packed, i.e. perfect
             packing.
    :rtype: float
    """
    w, h = bbox_size(sizes, positions)
    area_bounding_box = w*h
    area_rectangles = sum(w*h for w, h in sizes)
    return area_rectangles/area_bounding_box


def overlapping(sizes, positions):
    """Return indices of overlapping rectangles, else ``None``.

    Mainly used for test cases.

    Example::

        >>> sizes = [(10, 10), (10, 10)]

        >>> positions = [(0, 0), (10, 10)]
        >>> overlapping(sizes, positions) is None
        True

        >>> positions = [(0, 0), (5, 5)]
        >>> overlapping(sizes, positions)
        (0, 1)

    :param sizes: List of rectangle sizes (width, height).
    :type sizes: List[Tuple[int, int]]

    :param positions: List of rectangle positions (x, y).
    :type positions: List[Tuple[int, int]]

    :return: Return indices (i, j) if i-th and j-th rectangle overlap
             (first case found).  Return ``None`` if no rectangles
             overlap.
    :rtype: Union[None, Tuple[int, int]]
    """
    cdef:
        long w1, h1, x1, y1
        long w2, h2, x2, y2
        bint disjoint_in_x, disjoint_in_y
        Py_ssize_t n = len(sizes)
    for i in range(n):
        w1, h1 = sizes[i]
        x1, y1 = positions[i]
        for j in range(n):
            if j == i:
                continue
            w2, h2 = sizes[j]
            x2, y2 = positions[j]
            disjoint_in_x = (x1 + w1 <= x2 or x2 + w2 <= x1)
            disjoint_in_y = (y1 + h1 <= y2 or y2 + h2 <= y1)
            if not (disjoint_in_x or disjoint_in_y):
                return (i, j)
    return None



# Copyright 2017 - 2020 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from libcpp cimport bool
from libcpp.vector cimport vector
from libcpp.map cimport map
from libcpp.pair cimport pair
from libc.stdint cimport uint8_t, uint64_t
from libcpp.string cimport string

ctypedef pair[int,int] intpair
ctypedef pair[intpair, int] intpairint
ctypedef map[intpair, int] edgemap
ctypedef map[int, vector[int]] chainmap

cdef class labeldict(dict):
    cdef list _label
    def __init__(self,*args,**kwargs):
        super(labeldict,self).__init__(args,**kwargs)
        self._label = []
    def __missing__(self,l):
        self[l] = k = len(self._label)
        self._label.append(l)
        return k
    def label(self,k):
        return self._label[k]

cdef extern from "<memory>" namespace "std":
    cdef cppclass shared_ptr[T]:
        shared_ptr() nogil
        void reset(T*)

    cdef cppclass unique_ptr[T]:
        unique_ptr() nogil

cdef extern from "<random>" namespace "std":
    cdef cppclass default_random_engine:
        pass

cdef extern from "../include/find_embedding/util.hpp" namespace "find_embedding":
    cppclass LocalInteraction:
        pass

    ctypedef shared_ptr[LocalInteraction] LocalInteractionPtr


    cppclass optional_parameters:
        optional_parameters()
        void seed(uint64_t)

        LocalInteractionPtr localInteractionPtr
        int max_no_improvement
        default_random_engine rng
        double timeout
        double max_beta
        int tries
        int verbose
        bint interactive
        int inner_rounds
        int max_fill
        int chainlength_patience
        bint return_overlap
        bint skip_initialization
        chainmap fixed_chains
        chainmap initial_chains
        chainmap restrict_chains
        int threads

cdef extern from "src/pyutil.hpp" namespace "":
    cppclass LocalInteractionPython(LocalInteraction):
        LocalInteractionPython()

    ctypedef void (*cython_callback)(void *, int, const string &)

    cppclass LocalInteractionLogger(LocalInteraction):
        LocalInteractionLogger(cython_callback, void *)

    void handle_exceptions()

cdef extern from "../include/find_embedding/graph.hpp" namespace "graph":
    cppclass input_graph:
        input_graph()
        void push_back(int,int)
        int num_nodes()
        void clear()

cdef extern from "../include/find_embedding/pathfinder.hpp" namespace "find_embedding":
    cppclass pathfinder_public_interface

cdef extern from "../include/find_embedding/embedding_problem.hpp" namespace "find_embedding":
    cpdef enum VARORDER:
        VARORDER_SHUFFLE = 0
        VARORDER_DFS = 1
        VARORDER_BFS = 2
        VARORDER_PFS = 3
        VARORDER_RPFS = 4
        VARORDER_KEEP = 5

    cppclass parameter_processor:
        int num_vars
        optional_parameters params

cdef extern from "../include/find_embedding/embedding.hpp" namespace "find_embedding":
    pass

cdef extern from "../include/find_embedding/chain.hpp" namespace "find_embedding":
    cppclass pathfinder_wrapper:
        pathfinder_wrapper() except +handle_exceptions
        parameter_processor pp
        unique_ptr[pathfinder_public_interface] pf
        pathfinder_wrapper(input_graph &, input_graph &, optional_parameters &)
        int heuristicEmbedding() except +handle_exceptions
        int num_vars()
        void get_chain(int, vector[int] &)
        void set_initial_chains(chainmap &)
        void quickPass(const vector[int] &, int, int, bool, bool, double) except +handle_exceptions
        void quickPass(VARORDER, int, int, bool, bool, double) except +handle_exceptions

    cppclass chain:
        chain(vector[int] &w, int l)
        inline int size() const 
        inline int count(const int q) const
        inline int get_link(const int x) const 
        inline void set_link(const int x, const int q)
        inline int drop_link(const int x)
        inline void set_root(const int q)
        inline void clear()
        inline void add_leaf(const int q, const int parent)
        inline int trim_branch(int q)
        inline int trim_leaf(int q)
        inline int parent(const int q) const
        inline int refcount(const int q) const
        void link_path(chain &other, int q, const vector [int] &parents)
        inline void diagnostic(char *last_op)


cdef class cppembedding:
    cdef vector[chain] chains
    cdef vector[int] qubit_weights
    def __cinit__(self, int num_vars, int num_qubits):
        pass

cdef extern from "../include/find_embedding/find_embedding.hpp" namespace "find_embedding":
    int findEmbedding(input_graph, input_graph, optional_parameters, vector[vector[int]]&) except +handle_exceptions




# Copyright 2024 D-Wave
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
r"""Methods provided for the generation of one or more disjoint embeddings. 
These methods sequentially generate disjoint embeddings of a source graph 
onto a target graph or provide supporting functionality.
"""
import warnings
import time
from time import perf_counter
from typing import Union, Optional, Callable, Iterable, List

import dwave_networkx as dnx
import networkx as nx
import numpy as np

from minorminer.subgraph import find_subgraph

from minorminer.utils.feasibility import (
    embedding_feasibility_filter,
    lattice_size_lower_bound,
)


def shuffle_graph(
    G: nx.Graph,
    seed: Optional[Union[int, np.random.RandomState, np.random.Generator]] = None,
) -> nx.Graph:
    """Shuffle the node and edge ordering of a networkx graph.

    For embedding methods that operate as a function of the node or edge
    ordering this can force diversification in the returned embeddings. Note
    that special orderings that encode graph structure or geometry may enhance
    embedder performance (shuffling may lead to longer search times).

    Args:
        G: A networkx graph
        seed: When provided, is used to shuffle the order of nodes and edges in
        the source and target graph. This can allow sampling from otherwise deterministic routines.

    Returns:
        nx.Graph: The same graph with modified node and edge ordering.
    """
    prng = np.random.default_rng(seed)
    nodes = list(G.nodes())
    prng.shuffle(nodes)
    edges = list(G.edges())
    prng.shuffle(edges)
    _G = nx.Graph()
    _G.add_nodes_from(nodes)
    _G.add_edges_from(edges)
    return _G


def embeddings_to_array(embs: list, node_order=None, as_ndarray=False):
    """Convert list of embeddings into an array of embedding values

    Args:
        embs: A list of embeddings, each list entry in the
            form of a dictionary with integer values.
        node_order: An iterable giving the ordering of
            variables in each row. When not provided variables are ordered to
            match the first embedding :code:``embs[0].keys()``. node_order
            can define any subset of the source graph nodes (embedding
            keys).
        as_ndarray: The list is cast to an ndarray, this requires that
            the embedding is 1 to 1, that the embedding values (nodes of
            a target graph) can be cast as numpy numeric types.

    Raises:
        ValueError: If ``embs`` is empty and ``node_order`` not provided.

    Returns:
        An embedding matrix; each row defines an embedding ordered
        by node_order.
    """
    if node_order is None:
        if len(embs) == 0:
            if as_ndarray:
                raise ValueError("shape of ndarray cannot be inferred")
            else:
                return []

        else:
            node_order = embs[0].keys()

    if as_ndarray:
        if len(embs) == 0:
            return np.empty(shape=(0, len(node_order)))
        else:
            return np.asarray([[emb[v] for v in node_order] for emb in embs])

    else:
        return [[emb[v] for v in node_order] for emb in embs]


def array_to_embeddings(
    embs: Iterable[list], node_order: Optional[Iterable] = None
) -> list[dict]:
    """Convert list of embedding lists (values) to dictionary

    Args:
        embs: An iterable of embeddings, each embedding is a list where values are
            chains in node order.
        node_order: An iterable giving the ordering of
            variables in each row. When not provided variables are ordered to
            match the first embedding :code:``embs[0].keys()``. node_order
            can define any subset of the source graph nodes (embedding
            keys).

    Returns:
        A list of embedding dictionaries
    """

    if node_order is None:
        node_order = range(len(embs[0]))

    if len(embs) == len(node_order):
        embs = [{node_order[idx]: v for idx, v in enumerate(embs)}]
    else:
        embs = [{node_order[idx]: v for idx, v in enumerate(emb)} for emb in embs]
    return embs


def find_multiple_embeddings(
    S: nx.Graph,
    T: nx.Graph,
    *,
    max_num_emb: Union[None, int, float] = 1,
    use_filter: bool = True,
    embedder: Optional[callable] = None,
    embedder_kwargs: Optional[dict] = None,
    one_to_iterable: bool = False,
    shuffle_all_graphs: bool = False,
    seed: Optional[Union[int, np.random.RandomState, np.random.Generator]] = None,
    timeout: float = float("Inf"),
) -> list:
    """Finds multiple disjoint embeddings of a source graph onto a target graph

    Uses a greedy strategy to deterministically find multiple disjoint
    embeddings of a source graph onto a target graph; nodes used are removed
    after each successful embedding.

    Embedding multiple times on a large graph can take significant time.
    It is recommended the user adjust both the timeout parameter and
    embedder_kwargs appropriately (incl. timeout), and also consider bounding
    the number of embeddings returned (with max_num_emb).

    Args:
        S: The source graph to embed.
        T: The target graph in which to embed.
        max_num_emb: Maximum number of embeddings to find.
            Defaults to 1, set to None to find the maximum possible
            number. Use of float('Inf') is deprecated and gives equivalent
            behaviour to None.
        use_filter: Specifies whether to check feasibility
            of embedding arguments independently of the embedder method. In some
            easy to embed cases use of a filter can slow down operation.
        embedder: Specifies the embedding search method,
            a callable taking ``S``, ``T``, as the first two parameters. Defaults to
            ``minorminer.subgraph.find_subgraph``. When the embedder fails it is
            expected to return an empty dictionary.
        embedder_kwargs: Additional arguments for the embedder
            beyond ``S`` and ``T``.
        one_to_iterable: Determines how embedding mappings are
            interpreted. Set to ``True`` to allow multiple target nodes to map to
            a single source node. Defaults to ``False`` for one-to-one embeddings.
        shuffle_all_graphs: If True, the tile-masked target graph and source graph are
            shuffled on each embedding attempt. Note that, if the embedder supports
            randomization this should be preferred by use of embedder_kwargs. If an
            embedder is a function of the node or edge order this allows
            diversification of the embeddings found.
        seed: seed for the ``numpy.random.Generator`` controlling shuffling (if
            invoked).
        timeout: Total time allowed across all embeddings in seconds. Time elapsed is
            checked before each call to embedder. The total runtime is thereby bounded
            by timeout plus the time required by one call to embedder.

    Returns:
        list: A list of disjoint embeddings. Each embedding follows the format
        dictated by the embedder. By default, each embedding defines a 1:1
        map from the source to the target graph as a dictionary without
        reusing target variables.
    """
    embs = []
    if timeout <= 0:
        return embs

    timeout_at = perf_counter() + timeout

    if embedder is None:
        embedder = find_subgraph
    if embedder_kwargs is None:
        embedder_kwargs = {}

    if shuffle_all_graphs is True:
        prng = np.random.default_rng(seed)
        _T = T.copy()
        _T = shuffle_graph(T, seed=prng)
    else:
        _T = T

    if max_num_emb == float("inf"):
        warnings.warn(
            'Use of Inf for "max_num_emb" has been deprecated in favor of "None".',
            DeprecationWarning,
            stacklevel=2,
        )
        max_num_emb = None

    if max_num_emb is None:
        max_num_emb = T.number_of_nodes() // S.number_of_nodes()
    else:
        max_num_emb = min(
            T.number_of_nodes() // S.number_of_nodes(), round(max_num_emb)
        )

    if shuffle_all_graphs:
        _S = shuffle_graph(S, seed=prng)
    else:
        _S = S

    for _ in range(max_num_emb):
        if (
            use_filter
            and embedding_feasibility_filter(_S, _T, not one_to_iterable) is False
        ):
            emb = {}
        else:
            if perf_counter() >= timeout_at:
                break
            else:
                emb = embedder(_S, _T, **embedder_kwargs)

        if len(emb) == 0:
            break
        elif max_num_emb > 1:
            if len(embs) == 0:
                _T = T.copy()
            if one_to_iterable:
                _T.remove_nodes_from(n for c in emb.values() for n in c)
            else:
                _T.remove_nodes_from(emb.values())

        embs.append(emb)
    return embs


def lattice_size(T: nx.Graph) -> int:
    """Determines the cellular (square) dimension of a dwave_networkx lattice

    The lattice size is the parameter ``m`` of a dwave_networkx graph, also
    called number of rows, or in the case of a chimera graph max(m,n). This
    upper bounds the ``sublattice_size`` for ``find_sublattice_embeddings``.

    Args:
        T: The target graph in which to embed. The graph must be of type
            zephyr, pegasus or chimera and constructed by dwave_networkx.
    Returns:
        int: The maximum possible size of a tile.
    """
    # Possible feature enhancement, determine a stronger upper bound akin
    # to lattice_size_lower_bound, accounting for defects, the
    # degree distribution and other simple properties.

    return max(T.graph.get("rows"), T.graph.get("columns"))


def _is_valid_embedding(emb: dict, S: dict, T: dict, one_to_iterable: bool = True):
    """Check embedding validity.

    With special handling of 1:1 mappings."""
    from minorminer.utils.diagnostic import is_valid_embedding

    if one_to_iterable:
        return is_valid_embedding(emb, S, T)
    else:
        return is_valid_embedding({k: (v,) for k, v in emb.items()}, S, T)


def _mapped_proposal(emb: dict, f: Callable, one_to_iterable: bool = True):
    if one_to_iterable:
        return {k: tuple(f(n) for n in c) for k, c in emb.items()}
    else:
        return {k: f(n) for k, n in emb.items()}


def _local_graph(T=None, T_family=None, T_kwargs=None):
    if T_family is None:
        if T is None:
            raise ValueError(
                "Either the target graph T, or its lattice specication T_family and T_kwargs, must be provided"
            )
        return T.copy()
    else:
        if T_kwargs is None:
            T_kwargs = {}
        if T is not None:
            T_kwargs.update({"edge_list": T.edges, "node_list": T.nodes})
        if T_family == "chimera":
            return dnx.chimera_graph(**T_kwargs)
        elif T_family == "pegasus":
            return dnx.pegasus_graph(**T_kwargs)
        elif T_family == "zephyr":
            return dnx.zephyr_graph(**T_kwargs)
        else:
            raise ValueError("Unsupported dwave_networkx graph")


def find_sublattice_embeddings(
    S: nx.Graph,
    T: Optional[nx.Graph] = None,
    *,
    tile: Optional[nx.Graph] = None,
    sublattice_size: Optional[int] = None,
    max_num_emb: int = 1,
    use_filter: bool = True,
    use_tile_embedding: Optional[bool] = None,
    tile_embedding: Optional[dict] = None,
    seed: Optional[Union[int, np.random.RandomState, np.random.Generator]] = None,
    embedder: Optional[callable] = None,
    embedder_kwargs: Optional[dict] = None,
    one_to_iterable: bool = False,
    shuffle_all_graphs: bool = False,
    shuffle_sublattice_order: bool = False,
    timeout: float = float("Inf"),
    T_family: Optional[str] = None,
    T_kwargs: Optional[dict] = None,
) -> list:
    """Searches for embeddings on sublattices of the target graph.

    Selects a sublattice (subgrid) of the graph T using dwave_networkx
    sublattice tiles and mappings approprate to the target topology, and
    attemps to embed. When successful the nodes are removed and the
    process is repeated until all sublattices are explored or sufficient
    embeddings are yielded.

    Embedding multiple times on a large graph can take significant time.
    It is recommended the user adjust ``embedder_kwargs`` appropriately such
    as timeout, and also consider bounding the number of embeddings
    returned (with ``max_num_emb``).

    See https://doi.org/10.3389/fcomp.2023.1238988 for examples of usage.

    Args:
        S: The source graph to embed.
        T: The target graph in which to embed. Either the graph should be of type
            chimera, pegasus or zephyr and generated by dwave_networkx, or the
            graph must be cast to one of these types using `T_family` and `T_kwargs`.
            If `T` is None, then `T` is generated as a defect free graph.
        tile: A graph constructed by dwave_networkx applied as a displaceable mask
            to the target graph ``T``. The embedding search is limited to the mask.
            If the tile is not provided, it is generated as a fully yielded square
            sublattice of ``T`` of the same type (chimera, pegasus or zephyr), with
            ``m=sublattice_size``. See minorminer sublattice mapping documentation
            for details on compatible combinations. For example, chimera tiles
            can be used in combination with pegasus and zephyr target graphs.
        sublattice_size: Parameterizes the tile when it is not provided
           as an input argument: defines the number of rows and columns
           of a square sublattice (parameter m of the dwave_networkx graph
           family matching T). ``lattice_size_lower_bound()``
           provides a lower bound based on a fast feasibility filter. Note that
           the tile is created with default graph parameters (integer labeled variables).
        max_num_emb: Maximum number of embeddings to find.
            Defaults to 1, set to None for unbounded (try unlimited search on
            all lattice offsets).
        use_filter: Specifies whether to check feasibility of arguments for
            embedding independently of the embedder routine.  In some
            easy to embed cases use of a filter can slow down operation.
        use_tile_embedding: A single embedding for a tile is generated, and then
            reused for a maximum number of sublattices. If, due to defects or
            overlap with an existing assignment, the embedding fails at a particular
            offset embedder is not invoked.
            By default the setting is False unless ``tile==S`` or a tile_embedding
            is provided.
        tile_embedding: When provided, this should be an embedding from the source
            to the tile. If not provided it is generated by the embedder method, or
            fixed as a 1:1 mapping (in the special case that S==tile). Note
            that `one_to_iterable` should be adjusted to match the value type.
        embedder: Specifies the embedding search method, a callable taking ``S``, ``T`` as
            the first two arguments. Defaults to minorminer.subgraph.find_subgraph. Note
            that if `one_to_iterable` should be adjusted to match the return type.
            When the embedder fails it is expected to return an empty dictionary.
        embedder_kwargs: Dictionary specifying arguments for the embedder
            other than ``S``, ``T``.
        one_to_iterable: Specifies whether the embedder returns (and/or tile_embedding is)
            a dict with iterable values (True), or a 1:1 mapping (False).
        shuffle_all_graphs: If True, the tile-masked target graph and source graph are
            shuffled on each embedding attempt. Note that, if the embedder supports
            randomization this should be preferred by use of embedder_kwargs. If an
            embedder is a function of the node or edge order this allows
            diversification of the embeddings found.
        shuffle_sublattice_order: If True the ordering of sublattices (displacement
            of the tile (mask) on the target graph is randomized. This can allow
            for diversification of the embeddings found.
        seed: seed for the `numpy.random.Generator` controlling shuffling (if
            invoked).
        timeout: Total time allowed across all embeddings in seconds. Time elapsed is
            checked before each sublattice search begins, and before each call of the
            embedder function. The total runtime is thereby bounded by timeout plus the
            time required by one call to embedder.
        T_family: 'chimera', 'pegasus' or 'zephyr' arguments result in the target graph
            being cast to the corresponding dwave_networkx graph family.
        T_kwargs: The arguments passed to the graph generator. It is necessary to specify the
            shape parameters m (rows) and perhaps n (columns) or t (tile).
            If T is not None, `T_kwargs` is updated with the nodelist and edgelist of T.
    Raises:
        ValueError: If the target graph ``T`` is not of type zephyr, pegasus, or chimera and
            T_family, T_kwargs are not specified.
            If the tile_embedding is incompatible with the source graph and tile.

    Returns:
        list: A list of disjoint embeddings.
    """
    if timeout <= 0:
        return []
    timeout_at = perf_counter() + timeout
    if sublattice_size is None and tile is None:
        return find_multiple_embeddings(
            S=S,
            T=T,
            max_num_emb=max_num_emb,
            seed=seed,
            embedder=embedder,
            embedder_kwargs=embedder_kwargs,
        )
    if T is not None and T.graph.get("family", T_family) not in {
        "chimera",
        "pegasus",
        "zephyr",
    }:
        raise ValueError(
            "If T is not a dwave_networkx graph, T_family must be specified"
        )
    if T is not None and max_num_emb == 1 and seed is None and T_family is None:
        _T = T  # Graph is unmodified.
    else:
        _T = _local_graph(T, T_family, T_kwargs)

    if use_filter and tile is None:
        feasibility_bound = lattice_size_lower_bound(
            S=S, T=_T, one_to_one=not one_to_iterable
        )
        if feasibility_bound is None or sublattice_size < feasibility_bound:
            warnings.warn("sublattice_size < lower bound: embeddings will be empty.")
            return []

    family = _T.graph.get("family")
    if family == "chimera":
        sublattice_mappings = dnx.chimera_sublattice_mappings
        t = _T.graph["tile"]
        if tile is None:
            tile = dnx.chimera_graph(m=sublattice_size, n=sublattice_size, t=t)
        elif (
            use_filter
            and embedding_feasibility_filter(S, tile, not one_to_iterable) is False
        ):
            warnings.warn("tile is infeasible: embeddings will be empty.")
            return []

    elif family == "pegasus":
        sublattice_mappings = dnx.pegasus_sublattice_mappings
        if tile is None:
            tile = dnx.pegasus_graph(m=sublattice_size)
        elif (
            use_filter
            and embedding_feasibility_filter(S, tile, not one_to_iterable) is False
        ):
            warnings.warn("tile is infeasible: embeddings will be empty.")
            return []

    elif family == "zephyr":
        sublattice_mappings = dnx.zephyr_sublattice_mappings
        t = _T.graph["tile"]
        if tile is None:
            tile = dnx.zephyr_graph(m=sublattice_size, t=t)
        elif (
            use_filter
            and embedding_feasibility_filter(S, tile, not one_to_iterable) is False
        ):
            warnings.warn("tile is infeasible: embeddings will be empty.")
            return []

    else:
        raise ValueError(
            "source graphs must a graph constructed by "
            "dwave_networkx as chimera, pegasus or zephyr type"
        )

    if tile_embedding is not None:
        if not _is_valid_embedding(tile_embedding, S, tile, one_to_iterable):
            raise ValueError("tile_embedding is invalid for S and tile")
        use_filter = False  # Unnecessary

    if use_tile_embedding is None:
        use_tile_embedding = tile == S or tile_embedding is not None  # Trivial 1:1

    if use_tile_embedding and tile_embedding is None and S == tile:
        if one_to_iterable:
            tile_embedding = {
                i: (i,) for i in tile.nodes
            }  # 1 to 1 mapping is sufficient
        else:
            tile_embedding = {i: i for i in tile.nodes}  # 1 to 1 mapping is sufficient
        use_filter = False  # Unnecessary

    if use_filter or (use_tile_embedding and tile_embedding is None):
        # Check defect-free tile embedding is viable

        # With respect to filter:
        # * This assumes that an embedder that returns on a graph G
        # will also return an embedding for any graph where G is
        # a subgraph. Can be violated given heuristic search
        defect_free_embs = find_multiple_embeddings(
            S,
            tile,
            max_num_emb=1,
            use_filter=False,
            seed=seed,
            embedder=embedder,
            embedder_kwargs=embedder_kwargs,
            timeout=timeout_at - perf_counter(),
        )
        if len(defect_free_embs) == 0:
            # If embedding is infeasible on the tile*, it will be infeasible
            # on all subgraphs thereof (assuming sufficient time was provided
            # in embedder_kwargs)
            return []

        if use_tile_embedding:
            tile_embedding = defect_free_embs[0]
            # Apply sufficient restriction on the tile
            if one_to_iterable:
                tile = tile.subgraph({v for c in tile_embedding.values() for v in c})
            else:
                tile = tile.subgraph({v for v in tile_embedding.values()})

    if max_num_emb is None:
        max_num_emb = _T.number_of_nodes() // S.number_of_nodes()
    else:
        max_num_emb = min(_T.number_of_nodes() // S.number_of_nodes(), max_num_emb)

    embs = []

    if shuffle_all_graphs or shuffle_sublattice_order:
        prng = np.random.default_rng(seed)
    if shuffle_sublattice_order:
        sublattice_iter = list(sublattice_mappings(tile, _T))
        prng.shuffle(sublattice_iter)
    else:
        sublattice_iter = sublattice_mappings(tile, _T)

    for f in sublattice_iter:
        if perf_counter() > timeout_at:
            break
        if use_tile_embedding:
            proposal = _mapped_proposal(tile_embedding, f, one_to_iterable)
            if _is_valid_embedding(proposal, S, _T, one_to_iterable):
                sub_embs = [proposal]
            else:
                sub_embs = []
        else:
            Tr = _T.subgraph([f(n) for n in tile])
            sub_embs = find_multiple_embeddings(
                S,
                Tr,
                max_num_emb=max_num_emb - len(embs),
                use_filter=use_filter,
                seed=seed,
                embedder=embedder,
                embedder_kwargs=embedder_kwargs,
                timeout=timeout_at - perf_counter(),
            )
        embs += sub_embs
        if len(embs) >= max_num_emb:
            break

        for emb in sub_embs:
            # A potential feature extension would be to generate many
            # overlapping embeddings and solve an independent set problem. This
            # may allow additional parallel embeddings.
            if one_to_iterable:
                _T.remove_nodes_from([v for c in emb.values() for v in c])
            else:
                _T.remove_nodes_from(emb.values())
    return embs



# Copyright 2022 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import numbers

cimport cython
import numpy as np

from cython.operator cimport preincrement as inc, dereference as deref
from libcpp.algorithm cimport lower_bound as cpplower_bound
from libcpp.unordered_map cimport unordered_map

import dimod

from dimod.cyqmbase.cyqmbase_float64 import _dtype, _index_dtype
from dimod.cyutilities cimport as_numpy_float
from dimod.cyvariables cimport cyVariables
from dimod.libcpp.abc cimport QuadraticModelBase as cppQuadraticModelBase
from dimod.libcpp.constrained_quadratic_model cimport Penalty as cppPenalty
from dimod.libcpp.vartypes cimport Vartype as cppVartype
from dimod.sampleset import as_samples
from dimod.serialization.fileview import (
    SpooledTemporaryFile, _BytesIO,
    read_header, write_header,
    IndicesSection, LinearSection, OffsetSection, NeighborhoodSection, QuadraticSection,
    )
from dimod.typing cimport Numeric, float64_t
from dimod.variables import Variables


EXPRESSION_MAGIC_PREFIX = b"DIMODEXPR"


cdef class _cyExpression:
    def __init__(self, cyConstrainedQuadraticModel parent):
        self.parent = parent

        self.dtype = _dtype
        self.index_dtype = _index_dtype

    def __repr__(self):
        vartypes = {v: self.vartype(v).name for v in self.variables}
        return f"{type(self).__name__}({self.linear}, {self.quadratic}, {self.offset}, {vartypes})"

    @property
    def num_interactions(self):
        return self.expression().num_interactions()

    @property
    def num_variables(self):
        return self.expression().num_variables()

    @property
    def offset(self):
        return as_numpy_float(self.expression().offset())

    @offset.setter
    def offset(self, bias_type offset):
        self.expression().set_offset(offset)

    @property
    def variables(self):
        # we could do yet another view, but that way lies madness
        # todo: raise better error for disconnected and test

        variables = self.expression().variables()  # todo: check this a reference?

        cdef Py_ssize_t vi
        cdef cyVariables out = Variables()
        for vi in range(variables.size()):
            out._append(self.parent.variables.at(variables[vi]))
        return out

    def add_linear(self, v, bias_type bias):
        self.expression().add_linear(self.parent.variables.index(v), bias)

    def add_quadratic(self, u, v, bias):
        cdef Py_ssize_t ui = self.parent.variables.index(u)
        cdef Py_ssize_t vi = self.parent.variables.index(v)

        if ui == vi:
            if self.parent.cppcqm.vartype(ui) == cppVartype.SPIN:
                raise ValueError(f"SPIN variables (e.g. {self.variables[ui]!r}) "
                                 "cannot have interactions with themselves"
                                 )
            if self.parent.cppcqm.vartype(ui) == cppVartype.BINARY:
                raise ValueError(f"BINARY variables (e.g. {self.variables[ui]!r}) "
                                 "cannot have interactions with themselves"
                                 )

        if not self.parent.REAL_INTERACTIONS:
            if self.parent.cppcqm.vartype(ui) == cppVartype.REAL:
                raise ValueError(
                    f"REAL variables (e.g. {self.variables[ui]!r}) "
                    "cannot have interactions"
                    )
            if self.parent.cppcqm.vartype(vi) == cppVartype.REAL:
                raise ValueError(
                    f"REAL variables (e.g. {self.variables[vi]!r}) "
                    "cannot have interactions"
                    )

        self.expression().add_quadratic(ui, vi, bias)

    def add_variable(self, *args, **kwargs):
        """Add a variable to the model. See :meth:`ConstrainedQuadraticModel.add_variable()`."""
        return self.parent.add_variable(*args, **kwargs)

    def degree(self, v):
        return self.expression().degree(self.parent.variables.index(v))

    @cython.boundscheck(False)
    @cython.wraparound(False)
    def _energies(self, const Numeric[:, ::1] samples, cyVariables labels):
        cdef cppExpression[bias_type, index_type]* expression = self.expression()

        cdef Py_ssize_t num_samples = samples.shape[0]
        cdef Py_ssize_t num_variables = samples.shape[1]

        # let's reindex, using the underlying variable order
        cdef index_type[:] reindex = np.empty(expression.num_variables(), dtype=self.index_dtype)

        for i in range(expression.num_variables()):
            reindex[i] = labels.index(self.parent.variables.at(expression.variables()[i]))

        # we could avoid the copy at this point by checking if it's sorted and
        # the same length of the sample array, but let's not for now

        # we could do this manually, but way better to let NumPy handle it
        cdef const Numeric[:, ::1] subsamples = np.ascontiguousarray(np.asarray(samples)[:, reindex])

        cdef float64_t[::1] energies = np.empty(num_samples, dtype=np.float64)
        cdef Py_ssize_t si
        if subsamples.shape[1]:
            for si in range(num_samples):
                # we cast to QuadraticModelBase so that we're using the underlying variable order
                # rather than parent's
                energies[si] = (<cppQuadraticModelBase[bias_type, index_type]*>expression).energy(&subsamples[si, 0])
        else:
            for si in range(num_samples):
                energies[si] = 0

        return energies

    def energies(self, samples_like):
        samples, labels = as_samples(samples_like, labels_type=Variables)

        # we need contiguous and unsigned. as_samples actually enforces contiguous
        # but no harm in double checking for some future-proofness
        samples = np.ascontiguousarray(
                samples,
                dtype=f'i{samples.dtype.itemsize}' if np.issubdtype(samples.dtype, np.unsignedinteger) else None,
                )

        try:
            return np.asarray(self._energies(samples, labels))
        except TypeError as err:
            if np.issubdtype(samples.dtype, np.floating) or np.issubdtype(samples.dtype, np.signedinteger):
                raise err
            raise ValueError(f"unsupported sample dtype: {samples.dtype.name}")

    cdef cppExpression[bias_type, index_type]* expression(self) except NULL:
        # Not implemented. To be overwritten by subclasses.
        raise NotImplementedError

    @cython.boundscheck(False)
    @cython.wraparound(False)
    def _iindices(self):
        expression = self.expression()

        cdef index_type[:] indices = np.empty(expression.num_variables(), dtype=self.index_dtype)

        it = expression.variables().const_begin()
        end = expression.variables().const_end()
        cdef Py_ssize_t vi = 0
        while it != end:
            indices[vi] = deref(it)
            vi += 1
            inc(it)

        return np.asarray(indices)

    @cython.boundscheck(False)
    @cython.wraparound(False)
    def _iindices_load(self, buff, Py_ssize_t num_variables, object dtype):
        expression = self.expression()

        if expression.num_variables():
            raise RuntimeError("indices can only be loaded into an empty expression")

        cdef const index_type[:] indices = np.frombuffer(buff[:dtype.itemsize*num_variables], dtype=dtype)
        for vi in range(num_variables):
            expression.add_linear(indices[vi], 0)

    @cython.boundscheck(False)
    @cython.wraparound(False)
    def _ilinear(self):
        expression = self.expression()

        cdef bias_type[:] ldata = np.empty(expression.num_variables(), dtype=self.dtype)
        cdef Py_ssize_t vi
        for vi in range(expression.num_variables()):
            ldata[vi] = (<cppQuadraticModelBase[bias_type, index_type]*>expression).linear(vi)
        return np.asarray(ldata)

    @cython.boundscheck(False)
    @cython.wraparound(False)
    def _ilinear_load(self, buff, Py_ssize_t num_variables, object dtype):
        expression = self.expression()

        if expression.num_variables() != num_variables:
            raise RuntimeError("num_variables must match expression.num_variables()")

        cdef const bias_type[:] ldata = np.frombuffer(buff[:dtype.itemsize*num_variables], dtype=dtype)
        for vi in range(num_variables):
            (<cppQuadraticModelBase[bias_type, index_type]*>expression).set_linear(vi, ldata[vi])

    @cython.boundscheck(False)
    @cython.wraparound(False)
    def _iquadratic(self):
        expression = self.expression()

        # We choose the field names to mirror the quadratic iterator
        dtype = np.dtype([('u', self.index_dtype), ('v', self.index_dtype), ('bias', self.dtype)],
                         align=False)
        quadratic = np.empty(expression.num_interactions(), dtype)
        cdef index_type[:] irow = quadratic["u"]
        cdef index_type[:] icol = quadratic["v"]
        cdef bias_type[:] qdata = quadratic["bias"]

        cdef cppQuadraticModelBase[bias_type, index_type].const_quadratic_iterator2 it
        it = (<cppQuadraticModelBase[bias_type, index_type]*>expression).cbegin_quadratic()
        for vi in range(qdata.shape[0]):
            irow[vi] = deref(it).u
            icol[vi] = deref(it).v
            qdata[vi] = deref(it).bias

            inc(it)

        return quadratic

    @cython.boundscheck(False)
    @cython.wraparound(False)
    def _iquadratic_load(self, buff, Py_ssize_t num_interactions):
        # Note: it's possible to get segfaults by using this method directly.
        # We could add more safety, but for now I think marking it private is good
        # enough.
        expression = self.expression()

        if not expression.is_linear():
            raise RuntimeError("quadratic biases can only be loaded into a linear model")

        dtype = np.dtype([('u', self.index_dtype), ('v', self.index_dtype), ('bias', self.dtype)],
                         align=False)
        quadratic = np.frombuffer(buff[:dtype.itemsize*num_interactions], dtype=dtype)
        cdef const index_type[:] irow = quadratic["u"]
        cdef const index_type[:] icol = quadratic["v"]
        cdef const bias_type[:] qdata = quadratic["bias"]

        for i in range(num_interactions):
            # we're traversing the lower triangle, so it's OK to use add_quadratic_back
            (<cppQuadraticModelBase[bias_type, index_type]*>expression).add_quadratic_back(
                irow[i], icol[i], qdata[i])

    def get_linear(self, v):
        return as_numpy_float(self.expression().linear(self.parent.variables.index(v)))

    def get_quadratic(self, u, v, default=None):
        cdef Py_ssize_t ui = self.parent.variables.index(u)
        cdef Py_ssize_t vi = self.parent.variables.index(v)

        if ui == vi:
            if self.parent.cppcqm.vartype(ui) == cppVartype.SPIN:
                raise ValueError(f"SPIN variables (e.g. {u!r}) "
                                 "cannot have interactions with themselves"
                                 )
            if self.parent.cppcqm.vartype(ui) == cppVartype.BINARY:
                raise ValueError(f"BINARY variables (e.g. {v!r}) "
                                 "cannot have interactions with themselves"
                                 )
        cdef bias_type bias
        try:
            bias = self.expression().quadratic_at(ui, vi)
        except IndexError:
            # out of range error is automatically converted to IndexError
            if default is None:
                raise ValueError(f"{u!r} and {v!r} have no interaction") from None
            bias = default
        return as_numpy_float(bias)

    def is_linear(self):
        return self.expression().is_linear()

    def iter_neighborhood(self, v):
        cdef cppExpression[bias_type, index_type]* expression = self.expression()
        cdef Py_ssize_t vi = self.parent.variables.index(v)

        it = expression.cbegin_neighborhood(vi)
        while it != expression.cend_neighborhood(vi):
            v = self.parent.variables.at(deref(it).v)
            yield v, as_numpy_float(deref(it).bias)
            inc(it)

    def iter_quadratic(self):
        cdef cppExpression[bias_type, index_type]* expression = self.expression()

        it = expression.cbegin_quadratic()
        while it != expression.cend_quadratic():
            u = self.parent.variables.at(deref(it).u)
            v = self.parent.variables.at(deref(it).v)
            yield u, v, as_numpy_float(deref(it).bias)
            inc(it)

    def lower_bound(self, v):
        return self.parent.lower_bound(v)

    def reduce_linear(self):
        # todo: migrate to mixin
        raise NotImplementedError

    def reduce_neighborhood(self):
        # todo: migrate to mixin
        raise NotImplementedError

    def reduce_quadratic(self):
        # todo: migrate to mixin
        raise NotImplementedError

    def remove_interaction(self, u, v):
        cdef Py_ssize_t ui = self.parent.variables.index(u)
        cdef Py_ssize_t vi = self.parent.variables.index(v)

        self.expression().remove_interaction(ui, vi)

    def remove_variable(self, v):
        cdef Py_ssize_t vi = self.parent.variables.index(v)
        self.expression().remove_variable(vi)

    def set_linear(self, v, bias_type bias):
        cdef Py_ssize_t vi = self.parent.variables.index(v)
        self.expression().set_linear(vi, bias)

    def set_quadratic(self, u, v, bias):
        raise NotImplementedError

    def _into_file(self, fp):
        """Copy the constraint data into the given file-like.

        See ConstrainedQuadraticModel.to_file() for a description of the file format.
        """
        # This is intended to look like a cut-down quadratic model

        from dimod.constrained.constrained import CQM_SERIALIZATION_VERSION

        data = dict(shape=self.shape,
                    dtype=self.dtype.name,
                    itype=self.index_dtype.name,
                    type=type(self).__name__,
                    )

        write_header(fp, EXPRESSION_MAGIC_PREFIX, data, version=CQM_SERIALIZATION_VERSION)

        # the indices of each variable in the parent model
        fp.write(IndicesSection(self).dumps())

        # offset
        fp.write(OffsetSection(self).dumps())

        # linear
        fp.write(LinearSection(self).dumps())

        fp.write(QuadraticSection(self).dumps())

    def _from_file(self, fp):
        expr = self.expression()

        if expr.num_variables():
            raise RuntimeError("._from_file() can only be called on an empty expression")

        if isinstance(fp, (bytes, bytearray, memoryview)):
            file_like = _BytesIO(fp)
        else:
            file_like = fp

        header_info = read_header(file_like, EXPRESSION_MAGIC_PREFIX)

        # we don't bother checking the header version under the assumption that
        # this method is called from CQM.from_file

        cdef Py_ssize_t num_variables = header_info.data['shape'][0]
        cdef Py_ssize_t num_interactions = header_info.data['shape'][1]
        dtype = np.dtype(header_info.data['dtype'])
        itype = np.dtype(header_info.data['itype'])

        # variable indices
        self._iindices_load(IndicesSection.load(file_like),
                            dtype=self.index_dtype, num_variables=num_variables)

        # offset
        self.offset += OffsetSection.load(file_like, dtype=dtype)

        # linear
        self._ilinear_load(LinearSection.load(file_like, dtype=dtype, num_variables=num_variables),
                           dtype=dtype, num_variables=num_variables)

        # quadratic
        self._iquadratic_load(QuadraticSection.load(file_like), num_interactions=num_interactions)

    def upper_bound(self, v):
        return self.parent.upper_bound(v)

    def vartype(self, v):
        return self.parent.vartype(v)


cdef class cyObjectiveView(_cyExpression):
    cdef cppExpression[bias_type, index_type]* expression(self) except NULL:
        return &(self.parent.cppcqm.objective)


cdef class cyConstraintView(_cyExpression):
    def __init__(self, cyConstrainedQuadraticModel parent, object label):
        super().__init__(parent)
        self.constraint_ptr = parent.cppcqm.constraint_weak_ptr(parent.constraint_labels.index(label))

    cdef cppConstraint[bias_type, index_type]* constraint(self) except NULL:
        if self.constraint_ptr.expired():
            raise RuntimeError("this constraint is no longer valid")
        return self.constraint_ptr.lock().get()

    cdef cppExpression[bias_type, index_type]* expression(self) except NULL:
        return self.constraint()

    def is_discrete(self):
        constraint = self.constraint()
        return constraint.marked_discrete() and constraint.is_onehot()

    def is_onehot(self):
        return self.constraint().is_onehot()

    def is_soft(self):
        return self.constraint().is_soft()

    def mark_discrete(self, bint marker = True):
        self.constraint().mark_discrete(marker)

    def penalty(self):
        """The penalty type for a soft constraint.

        Returns:
            If the constraint is soft, returns one of ``'linear'`` or ``'quadratic'``.
            If the constraint is hard, returns ``None``.

        """
        constraint = self.constraint()
        penalty = constraint.penalty()

        if not constraint.is_soft():
            return None
        elif penalty == cppPenalty.LINEAR:
            return "linear"
        elif penalty == cppPenalty.QUADRATIC:
            return "quadratic"
        elif penalty == cppPenalty.CONSTANT:
            # user should never see this, but might as well be future proof
            return "constant"
        else:
            raise RuntimeError("unexpected penalty")

    def set_weight(self, weight, penalty='linear'):
        """Set the weight of the constraint.

        Args:
            weight: Weight for a soft constraint.
                Must be a positive number. If ``None`` or
                ``float('inf')``, the constraint is hard.
                In feasible solutions, all the model's hard constraints
                must be met, while soft constraints might be violated to achieve
                overall good solutions.

            penalty: Penalty type for a soft constraint (a constraint with its
                ``weight`` parameter set). Supported values are ``'linear'`` and
                ``'quadratic'``.
                ``'quadratic'`` is supported for a constraint with binary
                variables only.

        """
        cdef bias_type _weight = float('inf') if weight is None else weight

        if _weight <= 0:
            raise ValueError("weight must be a positive number or None")

        cdef cppConstraint[bias_type, index_type]* constraint = self.constraint()

        cdef cppPenalty _penalty
        if penalty == 'linear':
           _penalty = cppPenalty.LINEAR
        elif penalty == 'quadratic':
            for i in range(constraint.num_variables()):
                vartype = self.parent.cppcqm.vartype(constraint.variables()[i])
                if vartype not in (cppVartype.BINARY, cppVartype.SPIN):
                    raise ValueError("quadratic penalty only allowed if the constraint has binary variables")
            _penalty = cppPenalty.QUADRATIC
        elif penalty == 'constant':
            raise ValueError('penalty should be "linear" or "quadratic"')
        else:
            raise ValueError('penalty should be "linear" or "quadratic"')

        constraint.set_weight(_weight)
        constraint.set_penalty(_penalty)

    def weight(self):
        """The weight of the constraint.

        If the constraint is hard, will be ``float('inf')``.
        """
        return self.constraint().weight()



# Copyright 2020 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from libcpp.vector cimport vector
from libcpp.pair cimport pair
from libc.stdint cimport uint8_t, uint32_t, uint64_t
ctypedef vector[size_t] nodes_t
ctypedef vector[vector[size_t]] embedding_t
ctypedef vector[pair[size_t,size_t]] edges_t

cdef extern from "../include/busclique/util.hpp" namespace "busclique":
    cdef cppclass pegasus_spec:
        size_y dim_y
        size_x dim_x
        size_t shore
        pegasus_spec(size_t, vector[uint8_t], vector[uint8_t], uint32_t)
        nodes_t fragment_nodes(nodes_t)

    cdef cppclass chimera_spec:
        size_y dim_y
        size_x dim_x
        size_t shore
        chimera_spec(size_t, size_t, uint8_t, uint32_t)
        nodes_t fragment_nodes(nodes_t)

    cdef cppclass zephyr_spec:
        size_y dim_y
        size_x dim_x
        size_t shore
        zephyr_spec(size_t, uint8_t, uint32_t)
        nodes_t fragment_nodes(nodes_t)

    cdef cppclass serialize_size_tag:
        serialize_size_tag()

    cdef cppclass serialize_write_tag:
        serialize_write_tag()
    

cdef extern from "../include/busclique/cell_cache.hpp" namespace "busclique":
    cdef cppclass cell_cache[T]:
        cell_cache(T, nodes_t, edges_t)

cdef extern from "../include/busclique/bundle_cache.hpp" namespace "busclique":
    cdef cppclass bundle_cache[T]:
        bundle_cache(cell_cache[T] &)

cdef extern from "../include/busclique/clique_cache.hpp" namespace "busclique":
    cdef cppclass clique_cache[T]:
        clique_cache(cell_cache[T] &, bundle_cache[T] &, size_t)
        clique_cache(cell_cache[T] &, bundle_cache[T] &, size_t, size_t)

    cdef cppclass clique_iterator[T]:
        clique_iterator(cell_cache[T] &, clique_cache[T] &)
        int next(embedding_t &)

cdef extern from "../include/busclique/topo_cache.hpp" namespace "busclique":
    cdef cppclass topo_cache[T]:
        T topo
        topo_cache(T, nodes_t &, edges_t &)
        size_t serialize[t](t, uint8_t *) const
        nodes_t fragment_nodes() const
        edges_t fragment_edges() const
        void set_mask_bound(uint32_t)

cdef extern from "../include/busclique/find_clique.hpp" namespace "busclique":
    int find_clique[T](topo_cache[T] &, size_t, embedding_t &)
    void best_cliques[T](topo_cache[T], vector[embedding_t] &, embedding_t &)
    int short_clique[T](T, nodes_t, edges_t, embedding_t &)

cdef extern from "../include/busclique/find_biclique.hpp" namespace "busclique":
    void best_bicliques[T](topo_cache[T], vector[pair[pair[size_t, size_t], embedding_t]] &)

cdef extern from "../include/busclique/coordinate_types.hpp" namespace "busclique":
    cdef cppclass size_y:
        pass
    cdef cppclass size_x:
        pass

    size_t coordinate_index(size_y)
    size_t coordinate_index(size_x)



# Copyright 2024 D-Wave Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

cimport cython

from libc.stdint cimport (
    int8_t, int16_t, int32_t, int64_t,
    uint8_t, uint16_t, uint32_t, uint64_t,
    )

# Follow NumPy
# https://github.com/numpy/numpy/blob/882611cf11e1925bd8f3a8d7aa00d74675a7601c/numpy/__init__.pxd#L73-L74
# By following their behavior but not cimporting NumPy directly we can avoid  requiring
# NumPy at build-time.
ctypedef float float32_t
ctypedef double float64_t

# cython.integral only has short, int, long, so we need our own to also support int8
cdef fused SignedInteger:
    int8_t
    int16_t
    int32_t
    int64_t

# cython.integral doesn't support unsigned at all.
cdef fused UnsignedInteger:
    uint8_t
    uint16_t
    uint32_t
    uint64_t

cdef fused Integer:
    SignedInteger
    UnsignedInteger

ctypedef fused Numeric:
    int8_t
    int16_t
    int32_t
    int64_t
    float32_t
    float64_t





cdef extern from "rpackcore.h":

    ctypedef struct Rectangle:
        long width
        long height
        long x
        long y
        size_t index
        long area
        bint wide
        bint rotated

    ctypedef struct Cell

    cdef long start_pos(Cell *cell) nogil

    ctypedef struct Region:
        Cell *row_cell_start
        Cell *col_cell_start
        Cell *col_cell

    ctypedef struct BBoxRestrictions:
        long min_width
        long max_width
        long min_height
        long max_height
        long max_area

    ctypedef struct CGrid "Grid":
        size_t size
        long width
        long height

    cdef:
        CGrid *grid_alloc(size_t size, long width, long height) nogil
        void grid_free(CGrid *grid) nogil
        void grid_clear(CGrid *self) nogil
        int grid_find_region(CGrid *grid, Rectangle *rectangle, Region *reg) nogil
        void grid_split(CGrid *self, Region *reg) nogil
        long grid_search_bbox(CGrid *grid, Rectangle *sizes, BBoxRestrictions *bbr) nogil



# Copyright 2024 D-Wave Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import os
import platform
import sys

from Cython.Build import cythonize
from setuptools import setup, Extension


# Change directories so this works when called from other locations.
# Useful in build systems that build from source.
setup_folder_loc = os.path.dirname(os.path.abspath(__file__))
os.chdir(setup_folder_loc)

DEBUG = '--debug' in sys.argv or '-g' in sys.argv or 'CPPDEBUG' in os.environ
SAFE_COORDS = 'SAFE_COORDS' in os.environ

extra_compile_args = []
extra_compile_args_glasgow = []  # specific to glasgow
extra_compile_args_minorminer = []  # specific to minorminer
if platform.system().lower() == "windows":
    extra_compile_args.extend([
        '/std:c++17',
        '/MT',
        '/EHsc',
        ])

    extra_compile_args_glasgow.extend([
        '/external:W4',
        '/external:I', 'external',
        '/external:I', 'external/glasgow-subgraph-solver',
        '/DUSE_PORTABLE_SNIPPETS_BUILTIN',
        ])

    extra_compile_args_minorminer.extend([
        '/DCPPDEBUG' if DEBUG else None,
        '/DSAFE_COORDS' if SAFE_COORDS else None,
        ])

else:  # Unix
    extra_compile_args.extend([
        '-std=c++17',
        '-Wall',
        '-Wno-format-security',
        '--O0' if DEBUG else '-Ofast',
        '-fomit-frame-pointer',
        '-fipa-pure-const' if DEBUG else None,
        '-g' if DEBUG else '-g1',
        ])

    extra_compile_args_glasgow.extend([
        '-isystem', 'external',
        '-isystem', 'external/glasgow-subgraph-solver',
        '-DUSE_PORTABLE_SNIPPETS_BUILTIN',
        ])

    extra_compile_args_minorminer.extend([
        '-fno-rtti',
        '-DSAFE_COORDS' if SAFE_COORDS else None,
        ])

# filter out any None or empty arguments
extra_compile_args = list(filter(None, extra_compile_args))
extra_compile_args_glasgow = list(filter(None, extra_compile_args_glasgow))
extra_compile_args_minorminer = list(filter(None, extra_compile_args_minorminer))

# this is a subset of the total source files, so we can't just use glob or similar
glasgow_cc = [
    '/'.join(['external/glasgow-subgraph-solver/gss', f])
    for f in [
        'clique.cc',
        'configuration.cc',
        'homomorphism.cc',
        'innards/proof.cc',
        'restarts.cc',
        'sip_decomposer.cc',
        'timeout.cc',
        'innards/cheap_all_different.cc',
        'innards/graph_traits.cc',
        'innards/homomorphism_domain.cc',
        'innards/homomorphism_model.cc',
        'innards/homomorphism_searcher.cc',
        'innards/homomorphism_traits.cc',
        'innards/lackey.cc',
        'innards/svo_bitset.cc',
        'innards/thread_utils.cc',
        'innards/watches.cc',
        'formats/input_graph.cc',
        'formats/graph_file_error.cc',
    ]
]

extensions = [
    Extension(
        name="minorminer._minorminer",
        sources=["./minorminer/_minorminer.pyx"],
        include_dirs=['', './include/', './include/find_embedding'],
        language='c++',
        extra_compile_args=extra_compile_args + extra_compile_args_minorminer,
    ),
    Extension(
        name="minorminer.busclique",
        sources=["./minorminer/busclique.pyx"],
        include_dirs=['', './include/', '.include/busclique'],
        language='c++',
        extra_compile_args=extra_compile_args + extra_compile_args_minorminer,
    ),
    Extension(
        name="minorminer.subgraph",
        sources=["./minorminer/subgraph.pyx"] + glasgow_cc,
        include_dirs=['', './include', './external',
                      './external/glasgow-subgraph-solver'
                      './external/glasgow-subgraph-solver/src'
                      './external/glasgow-subgraph-solver/gss'],
        library_dirs=['./include', './external', './external/glasgow-subgraph-solver/gss/formats'],
        language='c++',
        extra_compile_args=extra_compile_args + extra_compile_args_glasgow,
    ),
    Extension(
        name="minorminer._extern.rpack._core",
        sources=["./minorminer/_extern/rpack/_core.pyx",
                 "./minorminer/_extern/rpack/src/rpackcore.c"],
        include_dirs=["./minorminer/_extern/rpack/include"],
        language='c',
    ),
]

setup(
    ext_modules=cythonize(extensions),
    packages=['minorminer',
              'minorminer.layout',
              'minorminer.utils',
              'minorminer._extern',
              'minorminer._extern.rpack',
              ],
    include_package_data=True,
)



# Copyright 2022 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#
# This file excerpts from the Glasgow Subgraph Solver, license follows.
#
# Copyright (c) 2013-2021 Ciaran McCreesh
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is 
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# distutils: language = c++
# cython: language_level = 3

from libcpp cimport bool
from libcpp.vector cimport vector
from libcpp.map cimport map
from libcpp.pair cimport pair
from libc.stdint cimport uint8_t, uint64_t
from libcpp.string cimport string
import random as _random

cdef class _labeldict(dict):
    cdef list _label
    def __init__(self,*args,**kwargs):
        super(_labeldict,self).__init__(args,**kwargs)
        self._label = []
    def __missing__(self,l):
        self[l] = k = len(self._label)
        self._label.append(l)
        return k
    def shuffle(self, random):
        random.shuffle(self._label)
        for i, l in enumerate(self._label):
            self[l] = i
    def label(self,k):
        return self._label[k]


cdef extern from "<chrono>" namespace "std::chrono" nogil:
    cdef cppclass time_point[T]:
        pass
    cdef cppclass steady_clock:
        pass
    cdef cppclass seconds:
        seconds()
    cdef cppclass milliseconds:
        seconds()
    cdef seconds make_seconds "std::chrono::seconds"(int)
    cdef milliseconds make_milliseconds "std::chrono::milliseconds"(int)
    cdef time_point[steady_clock] steady_clock_now "std::chrono::steady_clock::now"()

cdef extern from "glasgow-subgraph-solver/gss/timeout.hh" namespace "gss" nogil:
    cppclass Timeout:
        Timeout(seconds)

cdef extern from "glasgow-subgraph-solver/gss/formats/input_graph.hh" nogil:
    cdef cppclass InputGraph:
        InputGraph(int, bool, bool)
        void add_edge(int, int)
        void add_directed_edge(int, int, string)
        void resize(int)
        void set_vertex_label(int, string)

cdef extern from "glasgow-subgraph-solver/gss/restarts.hh" namespace "gss" nogil:
    cdef cppclass RestartsSchedule:
        pass
        
    cdef unsigned long long default_luby_multiplier "gss::LubyRestartsSchedule::default_multiplier"
    cdef milliseconds default_timed_duration "gss::TimedRestartsSchedule::default_duration"
    cdef unsigned long long default_timed_backtracks "gss::TimedRestartsSchedule::default_minimum_backtracks"
    cdef double default_geometric_constant "gss::GeometricRestartsSchedule::default_initial_value"
    cdef double default_geometric_multiplier "gss::GeometricRestartsSchedule::default_multiplier"

cdef extern from "<memory>" namespace "std" nogil:
    cdef cppclass shared_ptr[T]:
        pass
    cdef cppclass unique_ptr[T]:
        pass
    cdef shared_ptr[Timeout] make_shared_timeout "std::make_shared<gss::Timeout>"(seconds)
    cdef shared_ptr[InputGraph] make_shared[InputGraph](int, bool, bool)
    cdef InputGraph deref "*"(shared_ptr[InputGraph])
    cdef unique_ptr[RestartsSchedule] make_no_restarts_schedule "std::make_unique<gss::NoRestartsSchedule>"()
    cdef unique_ptr[RestartsSchedule] make_luby_restarts_schedule "std::make_unique<gss::LubyRestartsSchedule>"(unsigned long long)
    cdef unique_ptr[RestartsSchedule] make_geometric_restarts_schedule "std::make_unique<gss::GeometricRestartsSchedule>"(double, double)
    cdef unique_ptr[RestartsSchedule] make_timed_restarts_schedule "std::make_unique<gss::TimedRestartsSchedule>"(milliseconds, unsigned long long)


cdef extern from "glasgow-subgraph-solver/gss/vertex_to_vertex_mapping.hh" namespace "gss" nogil:
    cdef cppclass VertexToVertexMapping:
        pass

cdef extern from "glasgow-subgraph-solver/gss/value_ordering.hh" namespace "gss" nogil:
    #this enum contains None
    cdef enum ValueOrdering 'gss::ValueOrdering':
        _VO_None 'gss::ValueOrdering::None'
        _VO_Biased 'gss::ValueOrdering::Biased'
        _VO_Degree 'gss::ValueOrdering::Degree'
        _VO_AntiDegree 'gss::ValueOrdering::AntiDegree'
        _VO_Random 'gss::ValueOrdering::Random'

cdef extern from "glasgow-subgraph-solver/gss/homomorphism.hh" namespace "gss" nogil:
    cdef enum Injectivity 'gss:Injectivity':
        _I_Injective 'gss::Injectivity::Injective'
        _I_LocallyInjective 'gss::Injectivity::LocallyInjective'
        _I_NonInjective 'gss::Injectivity::NonInjective'

    cppclass HomomorphismParams:
        shared_ptr[Timeout] timeout
        time_point[steady_clock] start_time
        unique_ptr[RestartsSchedule] restarts_schedule
        unsigned n_threads
        bool triggered_restarts
        bool delay_thread_creation
        ValueOrdering value_ordering_heuristic
        Injectivity injectivity
        bool clique_detection
        bool distance3
        bool k4
        bool no_supplementals
        int number_of_exact_path_graphs
        bool clique_size_constraints
        bool clique_size_constraints_on_supplementals
        bool no_nds

    cppclass HomomorphismResult:
        map[int, int] mapping

cdef extern from "glasgow-subgraph-solver/gss/sip_decomposer.hh" namespace "gss" nogil:
    cdef HomomorphismResult solve_sip_by_decomposition(InputGraph, InputGraph, HomomorphismParams) except+

_default_kwarg = object()
def _check_kwarg(kwargs, name, default):
    value = kwargs.pop(name)
    return default if value is _default_kwarg else value

def find_subgraph(
    source,
    target,
    timeout=0,
    parallel=False,
    node_labels=None,
    edge_labels=None,
    as_embedding=False,
    injectivity='injective',
    seed=None,
    threads=1,
    triggered_restarts=_default_kwarg,
    delay_thread_creation=_default_kwarg,
    restarts_policy=_default_kwarg,
    luby_constant=_default_kwarg,
    geometric_constant=_default_kwarg,
    geometric_multiplier=_default_kwarg,
    restart_interval=_default_kwarg,
    restart_minimum=_default_kwarg,
    value_ordering='biased',
    clique_detection=True,
    no_supplementals=False,
    no_nds=False,
    distance3=False,
    k4=False,
    n_exact_path_graphs=4,
    cliques=False,
    cliques_on_supplementals=False,
    ):
    """
    Use the Glasgow Subgraph Solver to find a subgraph isomorphism from source
    to target.
    
    For more information, see the Glasgow subgraph solver at
    
        https://github.com/ciaranm/glasgow-subgraph-solver
        
    The full solver has more functionality than is exposed in this wrapper.
    
    If you use this function for research, please cite
    
        https://dblp.org/rec/conf/gg/McCreeshP020.html
        
    or if you use it in a non-research setting, please contact Ciaran McCreesh
    (ciaran.mccreesh@glasgow.ac.uk) if possible.  Their software is an output of
    taxpayer funded research, and it is very helpful for them if they can
    demonstrate real-world impact when they write grant applications.

    Args:
        S (iterable/NetworkX Graph): 
            The source graph as an iterable of node pairs representing the 
            edges, or a NetworkX Graph.

        T (iterable/NetworkX Graph):
            The target graph as an iterable of node pairs representing the 
            edges, or a NetworkX Graph.

        **params (optional): See below.

    Returns: 
        A dict that maps nodes in S to nodes in T. If no isomorphism is found,
        an empty dictionary is returned.

    Optional Parameters:
        timeout  (int, optional, default=0)
            Abort after this many seconds
        parallel (bool, optional, default=False):
            Use auto-configured parallel search (highly nondeterministic runtimes)
        node_labels (tuple, optional, default=None):
            If not ``None``, a pair of dicts (``S_labels``, ``T_labels``) whose keys are
            nodes and values are strings.  Unlabeled nodes are labeled with
            the empty string "".
        edge_labels (tuple, optional, default=None):
            If not ``None``, a pair of dicts (``S_labels``, ``T_labels``) whose keys are
            (source, dest) pairs of nodes corresponding to directed edges, and
            values are strings.  Unlabeled directed edges are labeled with the
            empty string "".  If the label on an edge (u, v) is intended to be
            undirected, you must provide the same label for both directions
            (u, v) and (v, u).
        as_embedding (bool, optional, default=False):
            If ``True``, the values of the returned dictionary will be singleton
            tuples similar to the return type of ``find_embedding``.
        injectivity (string, optional, default='injective'):
            Must be one of ('injective', 'locally injective', 'noninjective').
            By default, this function searches for subgraphs by finding injective
            homomorphisms.  That is, nodes of the target graph can only occur
            once in the output of the mapping.  By providing the default value
            'injective' to the `injectivity` parameter, that is true.  A mapping
            can be said to be 'locally injective' if the mapping is injective
            on the neighborhood of every node.
        seed (int/object, optional, default=None):
            If ``seed`` is an int, it will be used as a seed to a random number
            generator to randomize the algorithm.  This randomization is
            accomplished by shuffling the nodes and edges of the source and
            target graphs.  If ``seed`` is an object with an attribute named
            ``shuffle``, then that function will be called, with the expectation
            that it is equivalent to ``random.shuffle``.

            Note that the placement of nodes without incident edges is not
            subject to explicit randomization.

    Advanced Parallelism Options
        threads (int, optional, default=1):
            Use threaded search, with this many threads (0 to auto-detect)
            (this value is overridden to zero if ``paralell` is ``True``)
        triggered_restarts (bool, optional, default=False):
            Have one thread trigger restarts (more nondeterminism, better performance)
        delay_thread_creation (bool, optional, default=False):
            Do not create threads until after the first restart
            (default is changed to ``True`` if ``parallel`` is ``True``

    Advanced Search Configuration Options:
        restarts_policy (string, optional, default='luby'):
            Specify restart policy ('luby', 'geometric', 'timed' or 'none')
            (default policy is 'timed' with default parameters if ``parallel`` is ``True``)
        luby_constant (int, optional, default=666):
            Specify the starting constant / multiplier for Luby restarts
        geometric_constant (double, optional, 5400):
            Specify starting constant for geometric restarts
        geometric_multiplier (double, optional, default=1.0):
            Specify multiplier for geometric restarts
        restart_interval (int, optional, default=100):
            Specify the restart interval in milliseconds for timed restarts
        restart_minimum (int, optional, default=100):
            Specify a minimum number of backtracks before a timed restart can trigger    
        value_ordering (string, optional, default='biased'):
            Specify value-ordering heuristic ('biased', 'degree', 'antidegree', 'random', 'none')

    Advanced Input Processing Options:
        clique_detection (bool, optional, default=True):
            Enable clique / independent set detection
        no_supplementals (bool, optional, default=False):
            Do not use supplemental graphs
        no_nds (bool, optional, default=False):
            Do not use neighbourhood degree sequences

    Hidden Options:
        distance3 (bool, optional, default=False):
            Use distance 3 filtering (experimental)
        k4 (bool, optional, default=False):
            Use 4-clique filtering (experimental)
        n_exact_path_graphs (int, optional, default=4):
            Specify number of exact path graphs
        cliques (bool, optional, default=False):
            Use clique size constraints
        cliques_on_supplementals (bool, optional, default=False):
            Use clique size constraints on supplemental graphs too

    """
    if node_labels is None:
        node_labels = (None, None)
    if edge_labels is None:
        edge_labels = (None, None)

    cdef shared_ptr[InputGraph] source_g = make_shared[InputGraph](0, node_labels[0], edge_labels[0])
    cdef shared_ptr[InputGraph] target_g = make_shared[InputGraph](0, node_labels[1], edge_labels[1])

    if seed is None:
        random = None
    elif not hasattr(seed, 'shuffle'):
        random = _random.Random(seed)
    else:
        random = seed

    cdef _labeldict source_labels
    cdef _labeldict target_labels
    source_labels, source_isolated = _read_graph(deref(source_g), source, node_labels[0], edge_labels[0], random)
    target_labels, target_isolated  = _read_graph(deref(target_g), target, node_labels[1], edge_labels[1], random)
    cdef HomomorphismParams params

    if triggered_restarts is _default_kwarg:
        triggered_restarts = parallel

    check_kwargs = {
        'luby_constant': luby_constant,
        'geometric_constant': geometric_constant,
        'geometric_multiplier': geometric_multiplier,
        'restart_interval': restart_interval,
        'restart_minimum': restart_minimum,
    }
    if restarts_policy == 'luby':
        multiplier = _check_kwarg(check_kwargs, 'luby_constant', default_luby_multiplier)
        params.restarts_schedule = make_luby_restarts_schedule(multiplier)
    elif restarts_policy == 'geometric':
        constant = _check_kwarg(check_kwargs, 'geometric_constant', default_geometric_constant)
        multiplier = _check_kwarg(check_kwargs, 'geometric_multiplier', default_geometric_multiplier)
        params.restarts_schedule = make_geometric_restarts_schedule(constant, multiplier)
    elif restarts_policy == 'timed':
        interval = _check_kwarg(check_kwargs, 'restart_interval', None)
        backtracks = _check_kwarg(check_kwargs, 'restart_minimum', default_timed_backtracks)
        params.restarts_schedule = make_timed_restarts_schedule(
            default_timed_duration if interval is None else make_milliseconds(interval),
            backtracks
        )
    elif restarts_policy == 'none':
        params.restarts_schedule = make_no_restarts_schedule()
    elif restarts_policy is _default_kwarg:
        if parallel:
            params.restarts_schedule = make_timed_restarts_schedule(default_timed_duration, default_timed_backtracks)
        else:
            params.restarts_schedule = make_luby_restarts_schedule(default_luby_multiplier)                
    else:
        raise ValueError(f"restarts_policy {restarts_policy} not recognized")

    params.n_threads = 0 if parallel else threads

    if delay_thread_creation is _default_kwarg:
        delay_thread_creation = parallel
    params.delay_thread_creation = delay_thread_creation

    if value_ordering == 'none':
        params.value_ordering_heuristic = _VO_None
    elif value_ordering == 'biased':
        params.value_ordering_heuristic = _VO_Biased
    elif value_ordering == 'degree':
        params.value_ordering_heuristic = _VO_Degree
    elif value_ordering == 'antidegree':
        params.value_ordering_heuristic = _VO_AntiDegree
    elif value_ordering == 'random':
        params.value_ordering_heuristic = _VO_Random
    else:
        raise RuntimeError("unknown value ordering heuristic")

    if injectivity == 'injective':
        params.injectivity = _I_Injective
    elif injectivity == 'locally injective':
        params.injectivity = _I_LocallyInjective
    elif injectivity == 'noninjective':
        params.injectivity = _I_NonInjective
    else:
        raise RuntimeError("unrecognized injectivity option")

    if clique_detection is not _default_kwarg:
        params.clique_detection = clique_detection
    if distance3 is not _default_kwarg:
        params.distance3 = distance3
    if k4 is not _default_kwarg:
        params.k4 = k4
    if n_exact_path_graphs is not _default_kwarg:
        params.number_of_exact_path_graphs = n_exact_path_graphs
    if n_exact_path_graphs is not _default_kwarg:
        params.no_supplementals = no_supplementals
    if no_nds is not _default_kwarg:
        params.no_nds = no_nds
    if cliques is not _default_kwarg:
        params.clique_size_constraints = cliques
    if cliques_on_supplementals is not _default_kwarg:
        params.clique_size_constraints_on_supplementals = cliques_on_supplementals

    params.timeout = make_shared_timeout(make_seconds(timeout))
    params.start_time = steady_clock_now()

    check_kwargs = {k: v for k, v in check_kwargs.items() if v is not _default_kwarg}
    if check_kwargs:
        raise ValueError(f"unused parameters: {list(check_kwargs.keys())}")

    cdef HomomorphismResult result;
    if len(source_labels) + len(source_isolated) <= len(target_labels) + len(target_isolated) or injectivity != 'injective':
        result = solve_sip_by_decomposition(deref(source_g), deref(target_g), params)
        emb = dict((source_labels.label(s), target_labels.label(t)) for s, t in result.mapping)
    else:
        emb = {}

    if source_isolated and (len(emb) == len(source_labels)):
        if injectivity == 'injective':
            target_isolated.extend(set(target_labels)-set(emb.values()))
            if len(source_isolated) <= len(target_isolated):
                for s, t in zip(source_isolated, target_isolated):
                    emb[s] = t
        elif target_isolated or target_labels:
            t = next(iter(target_isolated or target_labels))
            for s in source_isolated:
                emb[s] = t

    if as_embedding:
        emb = {k: (v,) for k, v in emb.items()}

    return emb

cdef _read_graph(InputGraph &g, E, node_labels, edge_labels, random):
    cdef _labeldict L = _labeldict()
    cdef str label
    isolated_nodes = []
    if hasattr(E, 'edges'):
        G = E
        E = list(E.edges())
        for a, d in G.degree():
            if d:
                L[a]
            else:
                isolated_nodes.append(a)
    else:
        G = None

    if random is not None or node_labels is not None:
        if G is None:
            # E might be a generator... this is a silly-looking line but it
            # walks over the edge-list, puts every node into L, and leaves E
            # functionally unchanged
            E = [(L[a], L[b]) and (a, b) for a, b in E]

    if random is not None:
        L.shuffle(random)
        random.shuffle(E)

    if node_labels is not None:
        # performance note: we really wanna do this in order because as of
        # writing, the Glasgow implementation of set_vertex_label uses vector
        # erase/insert which can result in accidentally-quadratic runtime if
        # we don't write at the end
        for i, a in enumerate(L._label):
            label = node_labels.get(a)
            if label is not None:
                g.resize(i+1)
                g.set_vertex_label(i, bytes(label, "utf8"))

    if edge_labels is None:
        for a, b in E:
            g.add_edge(L[a],L[b])
    else:
        for a, b in E:
            label = edge_labels.get((a, b), "")
            g.add_directed_edge(L[a], L[b], bytes(label, "utf8"))
            label = edge_labels.get((b, a), "")
            g.add_directed_edge(L[b], L[a], bytes(label, "utf8"))

    g.resize(len(L))
    return L, isolated_nodes



# Copyright 2021 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

cimport cython

from dimod.typing cimport Numeric
from dimod.libcpp.binary_quadratic_model cimport BinaryQuadraticModel as cppBinaryQuadraticModel


cdef class cyBQM_template(cyQMBase):
    cdef cppBinaryQuadraticModel[bias_type, index_type]* cppbqm

    # developer note: we mostly use Py_ssize_t rather than size_t
    # since python does not really have an unsigned integer type that it
    # likes to use

    cdef Py_ssize_t _index(self, object, bint permissive=*) except -1
    cpdef Py_ssize_t add_linear_from_array(self, const Numeric[:] linear) except -1
    cpdef Py_ssize_t add_quadratic_from_dense(self, const Numeric[:, ::1] quadratic) except -1
    cpdef Py_ssize_t change_vartype(self, object) except -1
    cdef const cppBinaryQuadraticModel[bias_type, index_type]* data(self)
    cpdef Py_ssize_t resize(self, Py_ssize_t) except? 0
